import javax.swing.*;
import javax.swing.border.*;
import java.awt.*;
import java.awt.event.*;

/**
  * The FileDialog class is the set of dialog boxes used for displaying
  * currently saved games for saving and loading.
  *
  * @author  Joshua Thompson
  * @date    02/07/01
  * @version 1.1
  *
  */
//Version History
//  02/07/01 Joshua Thompson added java code
public class FileDialog extends JDialog implements ActionListener, KeyListener
{
    /**
     * The agreement button for the save game dialog.
     */
    private JButton saveOK;
    /**
     * The agreement button for the load game dialog.
     */
    private JButton loadOK;
    /**
     * The cancel button for both save and load game dialogs.
     */
    private JButton cancel;
    /**
     * The text field for the save game description.
     */
    private JTextField newSaveDesc;
    /**
     * An array of radio buttons for selecting a save game slot.
     */
    private JRadioButton[] slots;
    /**
     * The link to the Sokoban user interface.
     */
    private SGui owner;
    /**
     * The link to the Sokoban game interface.
     */
    private Game game;
    /**
     * A boolean variable that signifies if the load game was successful.
     */
    private boolean loaded;
    /**
     * A boolean variable that signifies if the save game was successful.
     */
    private boolean saved;

    /**
     * Constructor - initializes the member variables and sets up the keylistener.
     *
     * @param owner an instance of the Sokoban class for reference to the UI
     * @param game an instance of the Game class for reference to the game
     * interface
     */
    public FileDialog(SGui owner, Game game)
    {
        // SET this.owner = owner
        this.owner = owner;
        // SET this.game = game
        this.game = game;
        // ADD this as a key listener to the dialog box
        this.addKeyListener(this);
        // SET modal = true
        this.setModal(true);
        // SET loaded = false
        loaded = false;
        // SET saved = false
        saved = false;
        // INIT new JButton saveOK
        saveOK = new JButton("Save");
        // SET saveOK as default capable
        saveOK.setDefaultCapable(true);
        // SET saveOK mnemonic to S
        saveOK.setMnemonic(KeyEvent.VK_S);
        // ADD action listener this to saveOK
        saveOK.addActionListener(this);
        // INIT new JButton loadOK
        loadOK = new JButton("Load");
        // SET loadOK as default capable
        loadOK.setDefaultCapable(true);
        // SET loadOK mnemonic to L
        loadOK.setMnemonic(KeyEvent.VK_L);
        // ADD action listener this to loadOK
        loadOK.addActionListener(this);
        // INIT new JButton cancel
        cancel = new JButton("Cancel");
        // ADD action listener this to cancel
        cancel.addActionListener(this);
    }

    /**
     * The actionPerformed method responds to action events generated by the
     * buttons on the JDialog.  It is required for implementing ActionListener
     *
     * @param e the action event sent by the OS when an action is fired
     *
     * <br><b>preconditions:</b> an action has been fired
     * <br><b>postconditions:</b> the ActionEvent has been handled
     */
    public void actionPerformed(ActionEvent ev)
    {
        // IF e.getSource() = saveOK
        if (ev.getSource() == saveOK)
        {
            // GET selected save slot
            // INIT new int saveSlot = -1
            int saveSlot = -1;
            // INIT new int index = 0
            int index = 0;
            // INIT new boolean found = false
            boolean found = false;
            // WHILE not found and index < slots length DO
            while (!found && index < slots.length-1)
            {
                // IF slots[index] is selected THEN
                if (slots[index].isSelected())
                {
                    // SET saveSlot = index
                    saveSlot = index;
                    // ENDIF
                }
                // INCREMENT index

                index++;
                // ENDWHILE
            }
            // IF saveSlot = -1 THEN

            if (saveSlot == -1)
            {
                // INFORM user that no game is selected
                JOptionPane.showMessageDialog(owner, "No game selected.", "Save Game", JOptionPane.WARNING_MESSAGE);
            }
            // ELSE
            else
            {
                // GET new save game description
                String saveName = newSaveDesc.getText();
                // IF save description = "" THEN
                if (saveName.equals(""))
                {
                    // CONFIRM user wants no save game name
                    int confirm = JOptionPane.showConfirmDialog(owner, "Are you sure you want no save game description?", "Save Game", JOptionPane.YES_NO_OPTION);
                    // IF yes THEN
                    if (confirm == JOptionPane.YES_OPTION)
                    {
                        // CALL doSaveGame()
                        doSaveGame(saveSlot, " No Description");
                        // ENDIF
                    }
                }
                // ELSE IF save description length > 80

                else if (saveName.length() > 80)
                {
                    // SET save description = first 80 chars
                    saveName = saveName.substring(0, 80);
                    // CALL doSaveGame()
                    doSaveGame(saveSlot, saveName);
                }
                // ELSE
                else
                {
                    // CALL doSaveGame()
                    doSaveGame(saveSlot, saveName);
                    // ENDIF
                }
                // ENDIF
            }
        }
        // ELSE IF action was triggered by loadOK THEN

        else if (ev.getSource() == loadOK)
        {
            // GET selected save slot
            // SET loadSlot = -1
            int loadSlot = -1;
            // SET index = 0
            int index = 0;
            // SET found = false
            boolean found = false;
            // WHILE not found and index < slots length
            while (!found && index < slots.length-1)
            {
                // IF slots[index] is selected THEN
                if (slots[index].isSelected())
                {
                    // SET loadSlot = index
                    loadSlot = index;
                    // ENDIF
                }
                // INCREMENT index

                index++;
                // ENDWHILE
            }
            // IF loadSlot = -1 THEN

            if (loadSlot == -1)
            {
                // INFORM user that no game was selected
                JOptionPane.showMessageDialog(owner, "No game selected.", "Load Game", JOptionPane.WARNING_MESSAGE);
            }
            // ELSE
            else
            {
                // CALL game.loadGame() passing loadSlot
                game.loadGame(loadSlot);
                // SET loaded = true
                loaded = true;
                // DISPOSE of JDialog window
                this.dispose();
                // ENDIF
            }
        }
        // ELSE IF action was triggered by the cancel button THEN

        else if (ev.getSource() == cancel)
        {
            // DISPOSE of the JDialog window
            this.dispose();
            // SET loaded = false
            loaded = false;
            // SET saved = false
            saved = false;
            // ENDIF
        }
    }

    /**
     * Returns whether or not the game was successfully saved.
     *
     * @return true if save successful, otherwise false
     */
    public boolean isSaved()
    {
        // RETURN saved
        return saved;
    }

    /**
     * Calls the Game.saveGame() method, sets saved = true and disposes of
     * the JDialog window.
     *
     * @param saveSlot the save game number
     * @param saveName the description of the saved game
     */
    private void doSaveGame(int saveSlot, String saveName)
    {
        // CALL method game.saveGame() passing saveslot and savename
        game.saveGame(saveSlot, saveName);
        // SET saved = true
        saved = true;
        // DISPOSE of the JDialog window
        this.dispose();
    }

    /**
     * Displays the save game dialog to the user.
     *
     * <b>Precondition:</b> none.<br>
     * <b>Postcondition:</b> the save game dialog is displayed.
     */
    public void showSaveGameDialog()
    {
        // SET window title = "Save Game"
        this.setTitle("Save Game");
        // INIT new JPanel saves
        JPanel saves = new JPanel(new BorderLayout());
        // INIT new String[] desc = value returned by game.getSaveDescriptions()
        String[] desc = game.getSaveDescriptions();
        // INIT new JPanel saveSlots
        JPanel saveSlots = new JPanel(new GridLayout(desc.length, 1));
        // INIT slots[] to length of desc[]
        slots = new JRadioButton[desc.length];
        // INIT new button group
        ButtonGroup group = new ButtonGroup();
        // FOR index = 0 to desc.length LOOP
        for (int index = 0; index < desc.length; index++)
        {
            // IF desc[index] = null THEN
            if (desc[index] == null)
            {
                // SET slots[index] = "Empty Slot"
                slots[index] = new JRadioButton(" Empty Slot "+(index+1));
            }
            // ELSE
            else
            {
                // SET slots[index] = desc[index]
                slots[index] = new JRadioButton(desc[index]);
                // ENDIF
            }
            // ADD slots[index] to the buttongroup

            group.add(slots[index]);
            // ADD slots[index] to saveSlots JPanel
            saveSlots.add(slots[index]);
        }
        // SET slots[0] selected
        slots[0].setSelected(true);
        // INIT new JScrollPane
        JScrollPane saveSlotsPane = new JScrollPane(saveSlots, JScrollPane.VERTICAL_SCROLLBAR_ALWAYS, JScrollPane.HORIZONTAL_SCROLLBAR_NEVER);
        // SET preferred size = 400, 300
        saveSlotsPane.setPreferredSize(new Dimension(400, 300));
        // ADD new jlabel to the saves jpanel
        saves.add(new JLabel("Select Game to Save:"), BorderLayout.NORTH);
        // ADD saveSlotsPane to the saves jpanel
        saves.add(saveSlotsPane, BorderLayout.CENTER);
        // INIT newSaveDesc
        newSaveDesc = new JTextField();
        // INIT new JPanel getSaveDesc
        JPanel getSaveDesc = new JPanel(new BorderLayout());
        // ADD new jlabel to getSaveDesc jpanel
        getSaveDesc.add(new JLabel("Enter New Description:"), BorderLayout.NORTH);
        // ADD the newSaveDesc to the getSaveDesc jpanel
        getSaveDesc.add(newSaveDesc, BorderLayout.SOUTH);
        // ADD getSaveDesc to the saves jpanel
        saves.add(getSaveDesc, BorderLayout.SOUTH);
        // SET border for saves jpanel to an empty border of 15 pixels
        saves.setBorder(new EmptyBorder(15, 15, 15, 15));
        // INIT new JPanel buttons
        JPanel buttons = new JPanel(new FlowLayout());
        // ADD saveOK button to buttons jpanel
        buttons.add(saveOK);
        // ADD cancel button to the buttons jpanel
        buttons.add(cancel);
        // INIT new JPanel top
        JPanel top = new JPanel(new BorderLayout());
        // ADD saves jpanel to the top jpanel
        top.add(saves, BorderLayout.CENTER);
        // ADD buttons jpanel to the top jpanel
        top.add(buttons, BorderLayout.SOUTH);
        // ADD the top jpanel to the JDialog content pane
        this.getContentPane().add(top);
        // CALL method this.pack()
        this.pack();
        // SET the default button = saveOK
        this.getRootPane().setDefaultButton(saveOK);
        // SHOW the JDialog window
        this.show();
    }

    /**
     * Returns whether or not the game was successfully loaded.
     *
     * @return true if load successful, otherwise false
     */
    public boolean isLoaded()
    {
        // RETURN loaded
        return loaded;
    }

    /**
     * Displays the load game dialog to the user.
     *
     * <b>Precondition:</b> none.<br>
     * <b>Postcondition:</b> the load game dialog is displayed.
     */
    public void showLoadGameDialog()
    {
        // SET window title = "Load Game"
        this.setTitle("Load Game");
        // INIT new JPanel saves
        JPanel saves = new JPanel(new BorderLayout());
        // INIT new String[] desc = value returned by game.getSaveDescriptions()
        String[] desc = Game.getSaveDescriptions();
        // INIT new JPanel saveSlots
        JPanel saveSlots = new JPanel(new GridLayout(desc.length, 1));
        // INIT slots[] = length of desc[]
        slots = new JRadioButton[desc.length];
        // INIT new button group
        ButtonGroup group = new ButtonGroup();
        // INIT new int totalSaves = 0
        int totalSaves = 0;
        // FOR index = 0 to desc.length LOOP
        for (int index = 0; index < desc.length; index++)
        {
            // INIT new JRadioButton slots[index]
            slots[index] = new JRadioButton();
            // ADD slots[index] to the button group
            group.add(slots[index]);
            // IF desc[index] != null THEN
            if (desc[index] != null)
            {
                // SET text of slots[index] = desc
                slots[index].setText(desc[index]);
                // ADD slots[index] to saveSlots JPanel
                saveSlots.add(slots[index]);
                // INCREMENT totalSaves
                totalSaves++;
                // ENDIF
            }
            // ENDFOR
        }
        // IF totalSaves <= 0 THEN

        if (totalSaves <= 0)
        {
            // ADD jlabel "no saved games" to saveSlots
            saveSlots.add(new JLabel(" No Saved Games"));
        }
        // ELSE
        else
        {
            // SET slots[0] selected
            slots[0].setSelected(true);
        }
        // INIT new JScrollPane loadSlotsPane
        JScrollPane loadSlotsPane = new JScrollPane(saveSlots, JScrollPane.VERTICAL_SCROLLBAR_ALWAYS, JScrollPane.HORIZONTAL_SCROLLBAR_NEVER);
        // SET loadSlotsPane preferred size = 400 x 300
        loadSlotsPane.setPreferredSize(new Dimension(400, 300));
        // ADD a title to the saves JPanel
        saves.add(new JLabel("Select Game to Load:"), BorderLayout.NORTH);
        // ADD loadSlotsPane to the saves JPanel
        saves.add(loadSlotsPane, BorderLayout.CENTER);
        // SET the saves border = new EmptyBorder of 15 pixels around
        saves.setBorder(new EmptyBorder(15, 15, 15, 15));
        // INIT new buttons JPanel
        JPanel buttons = new JPanel(new FlowLayout());
        // ADD loadOK to buttons jpanel
        buttons.add(loadOK);
        // ADD cancel to buttons jpanel
        buttons.add(cancel);
        // INIT new jpanel top
        JPanel top = new JPanel(new BorderLayout());
        // ADD saves to top jpanel
        top.add(saves, BorderLayout.CENTER);
        // ADD buttons to bottom of top jpanel
        top.add(buttons, BorderLayout.SOUTH);
        // ADD top to JDialog content pane
        this.getContentPane().add(top);
        // CALL method this.pack()
        this.pack();
        // SET loadOK as default button
        this.getRootPane().setDefaultButton(loadOK);
        // CALL method this.show()
        this.show();
    }

    /**
     * Responds to the ESCAPE key being pressed to close the window.
     *
     * <b>precondition:</b> the jdialog is visible.
     * <b>postcondition:</b> the jdialog is not visible.
     */
    public void keyPressed(KeyEvent ev)
    {
        // IF the escape key was pressed THEN
        if (ev.getKeyCode() == KeyEvent.VK_ESCAPE)
        {
            // SET visible = false
            this.setVisible(false);
            // SET loaded = false
            loaded = false;
            // SET saved = false
            saved = false;
            // ENDIF
        }
    }

    /**
     * Empty method: required by keylistener.
     */
    public void keyReleased(KeyEvent ev)
    {}

    /**
     * Empty method: required by keylistener.
     */
    public void keyTyped(KeyEvent ev)
    {}
}
