import javax.swing.*;
import javax.swing.table.*;
import javax.swing.event.*;
import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.util.*;

/**
 * The LevelEditorUI class is a graphical interface used by the LevelEditor
 * to allow the user to create custom levels for use in the game.
 *
 * @author 	Sean Pinto
 * @date	02/27/2002
 * @version 1.0
 *
 */
public class LevelEditorUI extends JPanel
{
    /**
     * The LevelEditor where changes will be made to the custom level.
     */
    private Level editor;
    /**
     * The JTable display which shows the current state of the level being edited.
     */
    private JTable table;
    /**
     * The radio button that selects an empty square to be placed on the level.
     */
    private JRadioButton empty = new JRadioButton(new ImageIcon(Constants.graphicsPath+"empty.gif"), true);
    /**
     * The radio button that selects an wall square to be placed on the level.
     */
    private JRadioButton wall = new JRadioButton(new ImageIcon(Constants.graphicsPath+"wall.gif"), false);
    /**
     * The radio button that selects an crate square to be placed on the level.
     */
    private JRadioButton crate = new JRadioButton(new ImageIcon(Constants.graphicsPath+"crate.gif"), false);
    /**
     * The radio button that selects an incinerator square to be placed on the level.
     */
    private JRadioButton incinerator = new JRadioButton(new ImageIcon(Constants.graphicsPath+"incinerator.gif"), false);
    /**
     * The radio button that selects an robot facing right square to be placed on the level.
     */
    private JRadioButton robot = new JRadioButton(new ImageIcon(Constants.graphicsPath+"robotright.gif"), false);
    /**
     * The collection of brushes to be used for changing the level's squares.
     */
    private ButtonGroup brushes = new ButtonGroup();
    /**
     * The total number of robots currently placed on the map.
     */
    private int totalRobots = 0;


    /**
     * Constructor - initializes the JTable and the main display jpanel.
     */
    public LevelEditorUI()
    {
        // SET background color to black
        this.setBackground(Color.black);
        // SET layout to borderlayout
        this.setLayout(new BorderLayout());
        // INIT new LevelEditor editor
        editor = new Level();
        // INIT new JTable table
        table = new JTable(editor);
        // CALL method table.setShowHorizontalLines() passing false
        table.setShowHorizontalLines(false);
        // CALL method table.setShowVerticalLines() passing false
        table.setShowVerticalLines(false);
        // CALL method table.setAutoCreateColumnsFromModel() passing false
        table.setAutoCreateColumnsFromModel(false);
        // CALL method table.setRowHeight() passing Constats.gSizeY
        table.setRowHeight(Constants.gSizeY);
        // CALL method table.setAutoResizeMode() passing JTable.AUTO_RESIZE_OFF)
        table.setAutoResizeMode(JTable.AUTO_RESIZE_OFF);
        // CALL method table.setIntercellSpacing() passing Dimension (0, 0)
        table.setIntercellSpacing(new Dimension(0, 0));
        // CALL method table.setBackground() passing Color.black
        table.setBackground(Color.black);
        // INIT new MouseMonitor mouse
        MouseMonitor mouse = new MouseMonitor();
        // CALL method table.addMouseMotionListener() passing mouse
        table.addMouseMotionListener(mouse);
        // CALL method table.addMouseListener() passing mouse
        table.addMouseListener(mouse);
        // CALL addColumns method
        addColumns();
        // CALL addSideBar method
        addSideBar();
        newLevel();
    }


    /**
     * Adds the columns to the main JTable and sets column attributes.
     *
     * <br><b>preconditions:</b> the JTable exists
     * <br><b>postconditions:</b> the columns have been added to the JTable
     */
    public void addColumns()
    {

        // FOR col = 0 to Constants.maxCol
        for (int col = 0; col < Constants.maxCol; col++)
        {
            // INIT new TableColumn column
            TableColumn column = new TableColumn(col);
            // CALL method column.setCellRenderer() passing a new CellRenderer
            column.setCellRenderer(new CellRenderer());
            // CALL method column.setResizable(false)
            column.setResizable(false);
            // CALL method column.setMaxWidth(Constants.gSizeX)
            column.setMaxWidth(Constants.gSizeX);
            // CALL method column.setMinWidth(Constants.gSizeX)
            column.setMinWidth(Constants.gSizeX);
            // CALL method column.setPreferredWidth(Constants.gSizeX)
            column.setPreferredWidth(Constants.gSizeX);
            // CALL method table.addColumn() passing column
            table.addColumn(column);
        } // END FOR
        // CALL method add() passing table and BorderLayout.CENTER
        this.add(table, BorderLayout.CENTER);
    }


    /**
     * Adds the sideBar to the main panel.
     *
     * <br><b>preconditions:</b> none.
     * <br><b>postconditions:</b> the sidebar is added to the main jpanel
     */
    public void addSideBar()
    {
        // INIT new JPanel sideBarTop with GridLayout(5, 1)
        JPanel sideBarTop = new JPanel(new GridLayout(5, 1));

        // ADD empty to sideBarTop
        sideBarTop.add(empty);
        // ADD wall to sideBarTop
        sideBarTop.add(wall);
        // ADD crate to sideBarTop
        sideBarTop.add(crate);
        // ADD incinerator to sideBarTop
        sideBarTop.add(incinerator);
        // ADD robot to sideBarTop
        sideBarTop.add(robot);
        // ADD empty to the brushes ButtonGroup
        brushes.add(empty);
        // ADD wall to the brushes ButtonGroup
        brushes.add(wall);
        // ADD crate to the brushes ButtonGroup
        brushes.add(crate);
        // ADD incinerator to the brushes ButtonGroup
        brushes.add(incinerator);
        // ADD robot to the brushes ButtonGroup
        brushes.add(robot);
        // SET selectedIcon for empty
        empty.setSelectedIcon(new ImageIcon(Constants.graphicsPath+"empty_sel.gif"));
        // SET selectedIcon for wall
        wall.setSelectedIcon(new ImageIcon(Constants.graphicsPath+"wall_sel.gif"));
        // SET selectedIcon for crate
        crate.setSelectedIcon(new ImageIcon(Constants.graphicsPath+"crate_sel.gif"));
        // SET selectedIcon for incinerator
        incinerator.setSelectedIcon(new ImageIcon(Constants.graphicsPath+"incinerator_sel.gif"));
        // SET selectedIcon for robot
        robot.setSelectedIcon(new ImageIcon(Constants.graphicsPath+"robot_sel.gif"));
        // INIT new JPanel sideBar
        JPanel sideBar = new JPanel(new BorderLayout());
        // ADD sideBarTop to sideBar at BorderLayout.NORTH
        sideBar.add(sideBarTop, BorderLayout.NORTH);
        // CALL method add() passing sideBar and BorderLayout.EAST
        this.add(sideBar, BorderLayout.EAST);
    }


    /**
     * Saves the level that is currently being written.
     *
     * @param levelName the name of the level to save.
     * <BR><B>Pre Condition</B>There is a level that is being written
     * <BR><B>Post Condition</b>The level is saved to an external file.
     */
    public void saveLevel(String levelName)
    {
        // INIT robotsSaved to 0
        int robotsSaved = 0;
        // Make a PrintWriter
        PrintWriter out;
        // Create a temp Square currentSquare
        Square currentSquare;


        try
        {
            // IF the levelName ends with .lvl
            if(levelName.endsWith(".lvl"))
            {
                // INIT PrintWriter with a new FileWriter passing it levelName
                out = new PrintWriter(new FileWriter(levelName));
            }

            // ELSE
            else
            {
                // Concatinate level extension to levelName
                levelName = levelName + Constants.levelExt;
                // INIT printWriter with a new FileWriter passing it levelName
                out = new PrintWriter(new FileWriter(levelName));
            }  // END IF

            // FOR row = 0 through maxRow
            for(int row=0; row<Constants.maxRow; row++)
            {

                // FOR col = 0 through maxCol
                for(int col=0; col<Constants.maxCol; col++)
                {
                    // SET current square to the value at row col
                    currentSquare = (Square) editor.getValueAt(row, col);

                    // IF currentSquare is empty
                    if (currentSquare == Square.EMPTY)
                    {
                        // Print a 0 and a space to the file
                        out.print(0+" ");
                    }

                    // ELSE IF currentSquare is a wall
                    else if (currentSquare == Square.WALL)
                    {
                        // Print a 1 and a space to the file
                        out.print(1+" ");
                    }

                    // ELSE IF currentSquare is a crate
                    else if (currentSquare == Square.CRATE)
                    {
                        // Print a 2 and a space to the file
                        out.print(2+" ");
                    }

                    // ELSE IF currentSquare is an incinerator
                    else if (currentSquare == Square.INCINERATOR)
                    {
                        // Print a 3 and a space to the file
                        out.print(3+" ");
                    }

                    // ELSE IF currentSquare is a robot
                    else if (currentSquare == Square.ROBOT)
                    {

                        // IF the robotsSaved is less than 4
                        if (robotsSaved < 4)
                        {
                            // Print a 4 plus the number of robotsSaved and
                            // a space
                            out.print((4+robotsSaved)+" ");
                            // Increment robotsSaved
                            robotsSaved++;
                        }  // END IF
                    }  // END IF
                }  // END FOR
                // Print a new line into the file

                out.println("");
            }  // END FOR
            // Close the PrintWriter
            out.close();
        }  // END TRY

        // CATCH IOException
        catch (IOException iOException)
        {}  // END CATCH
    }


    /**
     * Loads a level into the main editing window to be edited.
     *
     * @param levelName the name of the level to load.
     * <BR><B>Pre Condition</B>There is a level that has been previously written.
     * <BR><B>Post Condition</B>The level is loaded into the editing window.
     */
    public void loadLevel(String levelName)
    {
        // INIT robotsLoaded to 0
        int robotsLoaded = 0;
        // Make new BufferedReader
        BufferedReader in;
        // Create String input
        String input;
        // Create StringTokenizer token
        StringTokenizer token;
        // INIT row, col, and totalRobots to 0
        int row=0;
        int col=0;
        totalRobots = 0;

        // TRY
        try
        {

            // IF levelName ends with .lvl
            if (levelName.endsWith(".lvl"))
            {
                // INIT BufferedReader with a new FileReader passing it levelName
                in = new BufferedReader(new FileReader(levelName));
            }

            // ELSE
            else
            {
                // Concatinate level extension to levelName
                levelName = levelName + Constants.levelExt;
                // INIT BufferedReader with a new FileReader passing it levelName
                in = new BufferedReader(new FileReader(levelName));
            }  // END IF
            // Create temp integer to hold tokens

            int square;

            // WHILE BufferedReader.isReady() AND row is less than maxRow
            while(in.ready() && row<Constants.maxRow)
            {
                // READ in line
                input = in.readLine();
                // Create new StringTokenizer
                token = new StringTokenizer(input, " ");
                // INIT col to 0
                col = 0;

                // WHILE token.hasMoreElements AND col is less than maxCol
                while(token.hasMoreElements() && col<Constants.maxCol)
                {
                    // Parse token into in and set to square
                    square = Integer.parseInt(token.nextToken());

                    // IF square = 0
                    if (square == 0)
                    {
                        // CALL editor.setValueAt with empty, row, col
                        editor.setValueAt(Square.EMPTY, row, col);
                    }

                    // ELSE IF square = 1
                    else if (square == 1)
                    {
                        // CALL editor.setValueAt with wall, row, col
                        editor.setValueAt(Square.WALL, row, col);
                    }

                    // ELSE IF square = 2
                    else if (square == 2)
                    {
                        // CALL editor.setValueAt with crate, row, col
                        editor.setValueAt(Square.CRATE, row, col);
                    }

                    // ELSE IF square = 3
                    else if (square == 3)
                    {
                        // CALL editor.setValueAt with incinerator, row, col
                        editor.setValueAt(Square.INCINERATOR, row, col);
                    }

                    // ELSE IF square is 4, 5, 6, or 7
                    else if (square >= 4 && square <= 7)
                    {
                        // IF totalRobots is less than 4
                        if (totalRobots < 4)
                        {
                            // CALL editor.setValueAt with robot, row, col
                            editor.setValueAt(Square.ROBOT, row, col);
                            // Increment totalRobots
                            totalRobots++;
                        }  // END IF
                    }

                    // ELSE

                    else
                    {
                        // CALL editor.setValueAt with empty, row, col
                        editor.setValueAt(Square.EMPTY, row, col);
                    }  // END IF
                    // Increment col
                    col++;
                }  // END WHILE
                // Increment row
                row++;
            }  // END WHILE
            // CLOSE BufferedReader
            in.close();
        }  // END TRY

        // CATCH
        catch (IOException iOException)
        {
            System.out.println("There was an I/O Exception at " + iOException);
        }  // END CATCH
    }


    /**
     * Creates an empty level in the editing window.
     *
     * <BR><B>Pre Condition</B>None
     * <BR><B>Post Condition</B>There is an empty level in the editing window.
     *
     */
    public void newLevel()
    {
        // FOR columns = 0 to maxCol
        for(int column=0; column<Constants.maxCol; column++)
        {

            // FOR row = 0 to maxRow
            for(int row=0; row<Constants.maxRow; row++)
            {
                // CALL editor.setValueAt with empty, row, column
                editor.setValueAt(Square.EMPTY, row, column);
            }  // END FOR
        }  // END FOR
    }

    /**
     * Returns the current level being edited
     *
     * <BR><B>Pre Condition</B>None
     * <BR><B>Post Condition</B>None
     *
     */
    public Level getEditor()
    {
        // RETURN current level
        return editor;
    }


    /**
     * The MouseMonitor class is the mouse interface that supports painting
     * the level squares.
     *
     * @author 	Sean Pinto
     * @date	02/27/2002
     * @version 1.0
     */
    class MouseMonitor extends MouseInputAdapter
    {

        /**
         *  This function calls mousePressed to facilitate dragging
         *
         * <br><b>preconditions:</b> none.
        	 * <br><b>postconditions:</b> none
         */
        public void mouseDragged (MouseEvent event)
        {
            // CALL method mousePressed passing event
            mousePressed(event);
        }

        /**
         *	This function responds to a mouse press event from the interface
         *  This function is called from a single press or via the
         *	mouseDragged method
         *
         * <br><b>preconditions:</b> none.
        	 * <br><b>postconditions:</b> A single cell is painted for the level editor table
         */
        public void mousePressed (MouseEvent event)
        {
            // IF event's component = table
            if (event.getComponent() == table)
            {

                // Create integer row and set to table.rowAtPoint(event.getPoint())
                int row = table.rowAtPoint(event.getPoint());
                // Create integer col and set to table.columnAtPoint(event.getPoint())
                int col = table.columnAtPoint(event.getPoint());
                // IF row and col = -1
                if (row != -1 && col != -1)
                {
                    // Set square to table.getValueAt(row,col)
                    Square square = (Square)table.getValueAt(row, col);

                    // IF the empty is selected
                    if (empty.isSelected())
                    {

                        // IF square = Square.ROBOT
                        if (square.equals(Square.ROBOT))
                        {
                            // Decrement totalRobots
                            totalRobots--;
                        }  // END IF
                        // Set square to EMPTY
                        square = Square.EMPTY;
                        // CALL editor.setValueAt(square, row, col)
                        editor.setValueAt(square, row, col);
                    }  // END IF

                    // ELSE IF wall is selected
                    else if (wall.isSelected())
                    {

                        // IF square = Square.ROBOT
                        if (square.equals(Square.ROBOT))
                        {
                            // Decrement totalRobots
                            totalRobots--;
                        }  // END IF
                        // Set square to WALL
                        square = Square.WALL;
                        // CALL editor.setValueAt(square, row, col)
                        editor.setValueAt(square, row, col);
                    }  // END IF

                    // ELSE IF crate is selected
                    else if (crate.isSelected())
                    {
                        // IF square = Square.ROBOT
                        if (square.equals(Square.ROBOT))
                        {
                            // Decrement totalRobots
                            totalRobots--;
                        }  // END IF
                        // Set square to Square.CRATE
                        square = Square.CRATE;
                        // CALL editor.setValueAt(square, row, col)
                        editor.setValueAt(square, row, col);
                    }  // END IF


                    // ELSE IF incinerator is selected
                    else if (incinerator.isSelected())
                    {
                        // IF square = Square.ROBOT
                        if (square.equals(Square.ROBOT))
                        {
                            // Decrement totalRobots
                            totalRobots--;
                        }  // END IF
                        // Set square to Square.INCINERATOR
                        square = Square.INCINERATOR;
                        // CALL editor.setValueAt(square, row, col)
                        editor.setValueAt(square, row, col);
                    }  // END IF

                    // ELSE IF robot is selected
                    else if (robot.isSelected())
                    {
                        // IF square = Square.ROBOT
                        if (!square.equals(Square.ROBOT))
                        {
                            // IF totalRobots < 4
                            if (totalRobots < 4)
                            {
                                // Increment totalRobots
                                totalRobots++;
                                // Set square to Square.ROBOT
                                square = Square.ROBOT;
                                // CALL editor.setValueAt(square, row, col)
                                editor.setValueAt(square, row, col);
                            }

                            // ELSE
                            else
                            {
                                // Show error dialog
                                JOptionPane.showMessageDialog(null,
                                                              "There are already 4 robots on the map.",
                                                              "Too many Robots", 0);
                            }  //END IF
                        }  //END IF
                    }  //END IF
                    // CALL table.validate()

                    table.validate();
                    // CALL table.repaint()
                    table.repaint();
                }  // END IF
            }  //END IF
        }
    }
}

