import javax.swing.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.border.*;
import java.util.StringTokenizer;
import java.util.Vector;
import java.net.*;

/**
  * The MultiplayerDialog class is the set of dialog boxes displayed when
  * a multiplayer game is running.
  *
  * @author  Joshua Thompson
  * @date    02/01/01
  * @version 1.1
  *
  */
//Version History
//  02/01/01 Joshua Thompson added java code
public class MultiplayerDialog extends JDialog implements ActionListener, KeyListener
{
    /**
     * The link to the Sokoban game interface.
     */
    protected Game game;
    /**
     * The link to the Sokoban user interface.
     */
    protected SGui owner;
    /**
     * The text field input box for the user's name.
     */
    protected JTextField userName = new JTextField();
    /**
     * The text field input box that displays the host's ip
     */
    protected JTextField myIP = new JTextField();
    /**
     * The text field input box to get the ip to connect to.
     */
    protected JTextField connectIP = new JTextField();
    /**
     * The text field input box to get the port to connect to.
     */
    protected JTextField connectPort = new JTextField();
    /**
     * The name of the level to be played.
     */
    protected String level;
    /**
     * A string list of the levels available to be played.
     */
    protected String[] levels;
    /**
     * The JList that contains the list of levels.
     */
    protected JList levelList;
    /**
     * The JList that contains the list of users.
     */
    protected JList userListDisplay;
    /**
     * The JButton that starts the host game setup.
     */
    protected JButton hostGame;
    /**
     * The JButton that starts the join game.
     */
    protected JButton joinGame;
    /**
     * JDialog starts waiting for connections when clicked.
     */
    protected JButton startHost;
    /**
     * JDailog attempts to connect to server specified in connectData when clicked.
     */
    protected JButton startJoin;
    /**
     * Allows the user to start the game.
     */
    protected JButton startGame;
    /**
     * The JButton that displays the host game help.
     */
    protected JButton helpHost;
    /**
     * The JButton that displays the join game help.
     */
    protected JButton helpJoin;
    /**
     * Allows the user to stop connecting or hosting the new multiplayer game.
     */
    protected JButton cancel;
    /**
     * A Timer that waits to see if the client has connected; used by
     * the attempting to connect dialog.
     */
    protected Timer connectTimer;

    /**
     * Constructor - sets up and displays the multiplayer dialog box.
     */
    public MultiplayerDialog(SGui owner, Game game)
    {
        super(owner, "New Multiplayer Game", false);
        // SET this.game = game
        this.game = game;
        // SET this.owner = owner
        this.owner = owner;
        // SET the default close operation to do nothing
        this.setDefaultCloseOperation(JDialog.DO_NOTHING_ON_CLOSE);
        // SET dialog box size to 400 x 300
        this.setSize(400, 300);
        // ADD key listener MultiplayerDialog
        this.addKeyListener(this);
        // INIT the connectTimer
        connectTimer = new Timer(100, this);
        // INIT userListDisplay
        userListDisplay = new JList();
        userListDisplay.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);

        // INIT start host button
        hostGame = new JButton("Host a Game");
        hostGame.setMnemonic(KeyEvent.VK_H);
        hostGame.addActionListener(this);
        // INIT start join button
        joinGame = new JButton("Join a Game");
        joinGame.setMnemonic(KeyEvent.VK_J);
        joinGame.addActionListener(this);
        // INIT cancel button
        cancel = new JButton("Cancel");
        cancel.setMnemonic(KeyEvent.VK_C);
        cancel.addActionListener(this);
        // INIT new JPanel mainPanel
        JPanel mainPanel = new JPanel(new BorderLayout());
        // INIT new JPanel buttons
        JPanel buttons = new JPanel(new GridLayout(5, 1));
        buttons.add(hostGame);
        buttons.add(new JLabel(" "));
        buttons.add(joinGame);
        buttons.add(new JLabel(" "));
        buttons.add(cancel);
        buttons.setBorder(new EmptyBorder(50, 50, 50, 50));
        // ADD buttons jpanel to the mainPanel
        mainPanel.add(buttons, BorderLayout.CENTER);
        this.getContentPane().add(mainPanel);
        this.setVisible(true);
    }

    /**
     * The actionPerformed method responds to action events generated by the
     * buttons on the JDialog.  It is required for implementing ActionListener
     *
     * @param e the action event sent by the OS when an action is fired
     *
     * <br><b>preconditions:</b> an action has been fired
     * <br><b>postconditions:</b> the ActionEvent has been handled
     */
    public void actionPerformed(ActionEvent ev)
    {
        // IF the action was triggered by connectTimer THEN
        if (ev.getSource() == connectTimer)
        {
            // IF game.isClientConnected() THEN
            if (game.isClientConnected())
            {
                // STOP the connectTimer
                connectTimer.stop();
                // CALL method showWaitingDialog(false)
                showWaitingDialog(false);
                // ENDIF
            }
        }
        // ELSE IF the action was triggered by the helpHost button THEN

        else if (ev.getSource() == helpHost)
        {
            owner.showHelpDialog(Constants.helpHostFile);
        }
        // ELSE IF the action was triggered by the helpJoin button THEN
        else if (ev.getSource() == helpJoin)
        {
            owner.showHelpDialog(Constants.helpJoinFile);
        }
        // ELSE IF the action was triggered by the hostGame button THEN
        else if (ev.getSource() == hostGame)
        {
            // CALL method showHostDialog()
            showHostDialog();
            // SET focus on the userName field
            userName.requestFocus();
        }
        // ELSE IF the action was triggered by the joinGame button THEN
        else if (ev.getSource() == joinGame)
        {
            // CALL method showJoinDialog()
            showJoinDialog();
            userName.requestFocus();
        }
        // ELSE IF the action was triggered by the startHost button THEN
        else if (ev.getSource() == startHost)
        {
            // GET the port
            String portText = connectPort.getText();
            int port;
            if (portText.equals(""))
            {
                port = 4000;
            }
            else
            {
                port = Integer.parseInt(portText);
            }
            // IF the port is valid (1 <= port <= 65535) THEN
            if (port > 65535 || port < 1)
            {
                // SHOW dialog box saying invalid port number
                JOptionPane.showMessageDialog(this, "Invalid port number.  Must be between 1 and 65535", "Host Error", JOptionPane.WARNING_MESSAGE);
            }
            // ELSE
            else
            {
                // CALL method setUserName()
                setUserName(userName.getText()+"(Host)");
                // CALL method game.startServer() passing port and this
                game.startServer(port, this);
                // CALL method showWaitingDialog() passing true
                showWaitingDialog(true);
                // ENDIF
            }
        }
        // ELSE IF the action was triggered by the connect button THEN

        else if (ev.getSource() == startJoin)
        {
            // GET the connect ip
            String ip = connectIP.getText();
            String portText = connectPort.getText();
            // CALL method setUserName()
            setUserName(userName.getText());
            // GET the port
            int port;
            if (portText.equals(""))
            {
                port = 4000;
            }
            else
            {
                port = Integer.parseInt(portText);
            }
            // IF ip = "" THEN
            if (ip.equals(""))
            {
                // SHOW dialog box saying no ip was specified
                JOptionPane.showMessageDialog(this, "No IP address specified.", "Connect Error", JOptionPane.WARNING_MESSAGE);
            }
            // ELSE IF the port is valid (1 <= port <= 65535) THEN
            else if (port > 65535 || port < 1)
            {
                // SHOW dialog box saying invalid port number
                JOptionPane.showMessageDialog(this, "Invalid port number.", "Connect Error", JOptionPane.WARNING_MESSAGE);
            }
            // ELSE
            else
            {
                // CALL method setUserName()
                setUserName(userName.getText());
                // CALL method game.startClient() passing ip, port and this
                game.startClient(ip, port, this);
                // CALL method showAttemptDialog()
                showAttemptingDialog();
                // ENDIF
            }
        }
        // ELSE IF the action was triggered by the startGame button THEN

        else if (ev.getSource() == startGame)
        {
            // SET level = levels[levelList.getSelectedIndex()]
            level = levels[levelList.getSelectedIndex()];
            // CALL method game.startMap() passing level
            game.startMap(level);
            // SET this.setVisible() = false
            this.setVisible(false);
        }
        // ELSE IF the action was triggered by the cancel button THEN
        else if (ev.getSource() == cancel)
        {
            // CALL method cancelDialog()
            cancelDialog();
            // ENDIF
        }
        // REPAINT owner

        owner.validate();
        owner.repaint();
        // REPAINT JDialog
        this.validate();
        this.repaint();
    }

    /**
     * Responds to the ESCAPE key being pressed to close the window.
     *
     * <b>precondition:</b> the jdialog is visible.
     * <b>postcondition:</b> the jdialog is not visible.
     */
    public void keyPressed(KeyEvent ev)
    {
        if (ev.getKeyCode() == KeyEvent.VK_ESCAPE)
        {
            cancelDialog();
        }
    }

    /**
     * Empty method: required by keylistener.
     */
    public void keyReleased(KeyEvent ev)
    {}

    /**
     * Empty method: required by keylistener.
     */
    public void keyTyped(KeyEvent ev)
    {}

    /**
     * Stops the connectTimer, ends the game and hides the multiplayer
     * connection dialog.
     *
     * <br><b>preconditions:</b> the multiplayer dialog is visible
     * <br><b>postconditions:</b> the multiplayer dialog is closed and the
     * new game creation is stopped.
     */
    public void cancelDialog()
    {
        // STOP the connectTimer
        connectTimer.stop();
        // CALL method owner.endGame()
        owner.endGame();
        // CALL method this.setVisible() passing false
        this.setVisible(false);
    }

    public void showHostDialog()
    {
        this.getContentPane().removeAll();
        // SET JDialog title = "Host Setup"
        this.setTitle("Host Setup");
        // INIT the mainPanel
        JPanel mainPanel = new JPanel(new BorderLayout());
        // INIT the connectData panel
        JPanel connectData = new JPanel(new GridLayout(6, 1));
        // ADD the user name JLabel and the text field
        connectData.add(new JLabel("User Name:"));
        connectData.add(userName);
        // ADD your ip address JLabel
        connectData.add(new JLabel("Your IP Address:"));
        // SET myIP editable = false
        myIP.setEditable(false);
        // ADD connectData() to myIP
        connectData.add(myIP);
        try
        {
            // SET myIP to the ip address of the localhost
            myIP.setText(InetAddress.getLocalHost().getHostAddress());
        }
        catch (UnknownHostException exc)
        {
            //exc.printStackTrace();
        }
        // ADD the port JLabel and the text field

        connectData.add(new JLabel("Port (Optional):"));
        connectData.add(connectPort);
        // ADD keylistener to consume invalid chars
        connectPort.addKeyListener(new PortListener());
        // SET the preferred size  = 250 x 200
        connectData.setPreferredSize(new Dimension(200, 100));
        connectData.setBorder(new EmptyBorder(15, 0, 50, 15));
        // ADD the connectData panel to mainPanel
        mainPanel.add(connectData, BorderLayout.WEST);
        // INIT String[] levels = game.getLevelList()
        levels = game.getLevelList();
        // INIT levelList
        levelList = new JList(levels);
        levelList.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
        levelList.setSelectedIndex(0);
        // INIT new JScrollPane levelScroll
        JScrollPane levelScroll = new JScrollPane(levelList, JScrollPane.VERTICAL_SCROLLBAR_ALWAYS, JScrollPane.HORIZONTAL_SCROLLBAR_NEVER);
        levelScroll.setPreferredSize(new Dimension(120, 80));
        // INIT new JPanel levelPanel
        JPanel levelPanel = new JPanel(new BorderLayout());
        // ADD levelScroll to levelPanel
        levelPanel.add(new JLabel("Select Level:"), BorderLayout.NORTH);
        levelPanel.add(levelScroll, BorderLayout.CENTER);
        levelPanel.setBorder(new EmptyBorder(20, 15, 20, 0));
        // ADD levelPanel to the mainPanel
        mainPanel.add(levelPanel, BorderLayout.EAST);
        mainPanel.setBorder(new EmptyBorder(10, 20, 10, 20));
        // INIT new JPanel buttons
        JPanel buttons = new JPanel(new FlowLayout());
        // INIT startHost JButton
        startHost = new JButton("Start");
        startHost.setMnemonic(KeyEvent.VK_S);
        startHost.addActionListener(this);
        // INIT help JButton
        helpHost = new JButton("Help");
        helpHost.setMnemonic(KeyEvent.VK_H);
        helpHost.addActionListener(this);
        // ADD startHost to buttons
        buttons.add(startHost);
        // ADD cancel to buttons
        buttons.add(cancel);
        // ADD helpHost to buttons
        buttons.add(helpHost);
        // ADD buttons to mainPanel
        mainPanel.add(buttons, BorderLayout.SOUTH);
        this.getContentPane().add(mainPanel);
        this.setVisible(true);
    }

    public void showJoinDialog()
    {
        // REMOVE all components from the JDialog
        this.getContentPane().removeAll();
        // SET JDialog title = "Host Setup"
        this.setTitle("Join Game");
        // INIT the mainPanel
        JPanel mainPanel = new JPanel(new BorderLayout());
        // INIT the connectData panel
        JPanel connectData = new JPanel(new GridLayout(6, 1));
        // ADD the user name JLabel and the text field
        connectData.add(new JLabel("User Name:"));
        connectData.add(userName);
        // ADD the connect ip JLabel and text field
        connectData.add(new JLabel("Connect IP (Get this # from the Host):"));
        connectData.add(connectIP);
        // ADD the port JLabel and the text field
        connectData.add(new JLabel("Port (Optional):"));
        connectData.add(connectPort);
        // ADD keylistener to consume invalid chars
        connectPort.addKeyListener(new PortListener());
        // SET the preferred size  = 250 x 200
        connectData.setPreferredSize(new Dimension(200, 100));
        connectData.setBorder(new EmptyBorder(15, 0, 50, 15));
        // ADD the connectData jpanel to the mainPanel
        mainPanel.add(connectData, BorderLayout.CENTER);
        // INIT new JPanel buttons
        JPanel buttons = new JPanel(new FlowLayout());
        // INIT startHost JButton
        startJoin = new JButton("Join");
        startJoin.setMnemonic(KeyEvent.VK_J);
        startJoin.addActionListener(this);
        // INIT help JButton
        helpJoin = new JButton("Help");
        helpJoin.setMnemonic(KeyEvent.VK_H);
        helpJoin.addActionListener(this);
        // ADD startHost to buttons
        buttons.add(startJoin);
        // ADD cancel to buttons
        buttons.add(cancel);
        // ADD helpHost to buttons
        buttons.add(helpJoin);
        // ADD buttons to mainPanel
        mainPanel.add(buttons, BorderLayout.SOUTH);
        mainPanel.setBorder(new EmptyBorder(10, 20, 10, 20));
        this.getContentPane().add(mainPanel);
        this.setVisible(true);
    }

    private class PortListener extends KeyAdapter
    {
        public void keyTyped(KeyEvent ev)
        {
            if (!((ev.getKeyChar() >= '0' && ev.getKeyChar() <= '9')
                    || ev.getKeyCode() == KeyEvent.VK_DELETE
                    || ev.getKeyChar() == '\b'))
            {
                ev.consume();
            }
        }
    }

    /**
     * Displays the waiting for game to start dialog window.
     *
     * @param hosting boolean variable signifying if the user is hosting or not.
     *
     * <br><b>preconditions:</b> none.
     * <br><b>postconditions:</b> the waiting for game start dialog is displayed.
     */
    public void showWaitingDialog(boolean hosting)
    {
        // REMOVE all compononts from the content pane
        this.getContentPane().removeAll();
        // INIT mainPanel
        JPanel mainPanel = new JPanel(new BorderLayout());
        // SET mainPanel border = new EmptyBorder of 10 pixels around
        mainPanel.setBorder(new EmptyBorder(10, 10, 10, 10));
        // INIT new JScrollPane userScroll containing userListDisplay
        JScrollPane userScroll = new JScrollPane(userListDisplay, JScrollPane.VERTICAL_SCROLLBAR_ALWAYS, JScrollPane.HORIZONTAL_SCROLLBAR_NEVER);
        // SET the preferred size = 200 x 120
        userScroll.setPreferredSize(new Dimension(200, 120));
        // INIT new JPanel users
        JPanel users = new JPanel(new BorderLayout());
        // INIT new JLabel usersDesc
        JLabel usersDesc = new JLabel("Players Connected:");
        // SET usersDesc border = new Empty Border of 10 pixels on the bottom
        usersDesc.setBorder(new EmptyBorder(0, 0, 10, 0));
        // ADD userDesc to the top of the users jpanel
        users.add(usersDesc, BorderLayout.NORTH);
        // ADD the userScroll to the bottom of the users jpanel
        users.add(userScroll, BorderLayout.CENTER);
        // ADD the users jpanel to the mainPanel
        mainPanel.add(users, BorderLayout.WEST);
        // INIT new JPanel startButtons
        JPanel buttons = new JPanel(new FlowLayout());
        buttons.setBorder(new EmptyBorder(20, 0, 0, 0));
        // IF hosting THEN
        if (hosting)
        {
            // SET the window title
            this.setTitle("Waiting for Players");
            // INIT new JPanel ip displaying the user's ip address
            JPanel ip = new JPanel(new BorderLayout());
            ip.add(new JLabel("Your IP Address:  "), BorderLayout.WEST);
            ip.add(myIP, BorderLayout.CENTER);
            ip.setBorder(new EmptyBorder(0, 0, 5, 0));
            // ADD the ip jpanel to the mainPanel
            mainPanel.add(ip, BorderLayout.NORTH);
            // INIT startGame JButton
            startGame = new JButton("Start Game");
            startGame.setMnemonic(KeyEvent.VK_S);
            startGame.addActionListener(this);
            // ADD startGame button to startButtons jpanel
            buttons.add(startGame);
            // ADD cancel button to startButtons jpanel
            buttons.add(cancel);
        }
        // ELSE
        else
        {
            // SET the window title
            this.setTitle("Waiting for Host to Start Game");
            // SET text of cancel button = "Disconnect"
            cancel.setText("Leave");
            // SET cancel button mnemonic = 'L'
            cancel.setMnemonic(KeyEvent.VK_L);
            // ADD cancel button to startButtons jpanel
            buttons.add(cancel);
        }
        // ADD the startButtons jpanel to the bottom of the mainPanel
        mainPanel.add(buttons, BorderLayout.SOUTH);
        // ADD the mainPanel to the JDialog content pane
        this.getContentPane().add(mainPanel);
        // CALL method this.pack()
        this.pack();
    }

    /**
     * Shows a dialog informing user of attempting to connect to a multiplayer
     * game.
     *
     * <br><b>preconditions:</b> none
     * <br><b>postconditions:</b> the attempting to connect dialog is visible
     */
    private void showAttemptingDialog()
    {
        // START the connectTimer
        connectTimer.start();
        // REMOVE the components from the JDialog content pane
        this.getContentPane().removeAll();
        // INIT new mainPanel
        JPanel mainPanel = new JPanel(new BorderLayout());
        mainPanel.setBorder(new EmptyBorder(10, 10, 10, 10));
        // INIT new JLabel attemptMsg
        JLabel attemptMsg = new JLabel("Attempting to connect to host: "+connectIP.getText(), JLabel.CENTER);
        // ADD attemptMsg to the mainPanel
        mainPanel.add(attemptMsg, BorderLayout.CENTER);
        // INIT new JPanel buttons
        JPanel buttons = new JPanel(new FlowLayout());
        // ADD cancel button to buttons jpanel
        buttons.add(cancel);
        buttons.setBorder(new EmptyBorder(20, 10, 10, 10));
        // ADD buttons jpanel to the bottom of the mainPanel
        mainPanel.add(buttons, BorderLayout.SOUTH);
        // ADD the mainPanel to the JDialog content pane
        this.getContentPane().add(mainPanel);
        // CALL method this.pack()
        this.pack();
    }

    /**
     * Shows a confirmation dialog box asking if the user wishes to restart
     * the current level.
     *
     * @param userName the name of the user requesting the restart
     * @return true if the user selects yes, otherwise false
     * <br><b>preconditions:</b> none
     * <br><b>postconditions:</b> the request dialog has been displayed and the
     * user has selected yes or no.
     */
    public boolean showRestartRequestDialog(String userName)
    {
        // INIT new request String
        String request = userName+" wishes to restart the level.  Do you agree?";
        // SHOW confirm dialog box with request string
        int confirm = JOptionPane.showConfirmDialog(owner, request, "Restart Request", JOptionPane.YES_NO_OPTION, JOptionPane.PLAIN_MESSAGE);
        // IF yes THEN
        if (confirm == JOptionPane.YES_OPTION)
        {
            // RETURN true
            return true;
        }
        // ELSE
        else
        {
            // RETURN false
            return false;
            // ENDIF
        }
    }

    /**
     * Shows a confirmation dialog box asking if the user wishes to change
     * the current level.
     *
     * @param userName the name of the user requesting the level change
     * @param levelName the name of the level to change to
     * @return true if the user selects yes, otherwise false
     * <br><b>preconditions:</b> none
     * <br><b>postconditions:</b> the request dialog has been displayed and the
     * user has selected yes or no.
     */
    public boolean showChangeLevelRequestDialog(String userName, String levelName)
    {
        // INIT new request String
        String request = userName+" wishes to change levels to: "+levelName+".  Do you agree?";
        // SHOW confirm dialog box with request string
        int confirm = JOptionPane.showConfirmDialog(owner, request, "Change Level Request", JOptionPane.YES_NO_OPTION, JOptionPane.PLAIN_MESSAGE);
        // IF yes THEN
        if (confirm == JOptionPane.YES_OPTION)
        {
            // RETURN true
            return true;
        }
        // ELSE
        else
        {
            // RETURN false
            return false;
            // ENDIF
        }
    }

    /**
     * Gets the user name entered from the userName JTextField and then
     * passes the string to the game's setUserName() method.
     */
    private void setUserName(String inputname)
    {
        // INIT new String name = userName.getText()
        String name = inputname;
        // IF name = "" THEN
        if (name.equals(""))
        {
            // SET name = "Player"
            name = "Player";
        }
        // ELSE
        else
        {
            // REMOVE spaces
            StringTokenizer tkn = new StringTokenizer(name, " ");
            name = "";
            while(tkn.hasMoreTokens())
            {
                name = name + tkn.nextToken();
            }
            // ENDIF
        }
        // CALL method game.setUserName() passing name

        game.setUserName(name);
    }

    /**
     * Returns the name of the level selected in the levelList JList
     *
     * @return the name of the level selected in the level list display
     */
    public String getLevelName()
    {
        // RETURN level
        return level;
    }

    /**
     * Updates the userListDisplay with the current robots' user names.
     *
     * <br><b>preconditions:</b> the userListDisplay has been initialized
     * <br><b>postconditions:</b> the user list is updated
     */
    public void updateUserList(Robot[] robots)
    {
        // INIT new String[] userListData = robots.length
        String[] userListData = new String[robots.length];
        // FOR index = 0 to userListData.length
        for (int index = 0; index < userListData.length; index++)
        {
            // SET userListData[index] = robots[index].getUserName()
            userListData[index] = robots[index].getUserName();
            // ENDFOR
        }
        // CALL method userListDisplay.setListData() passing userListData

        userListDisplay.setListData(userListData);
    }

    /**
     * Shows a dialog informing user of that the connection with the server
     * has been closed.
     *
     * <br><b>preconditions:</b> none
     * <br><b>postconditions:</b> the connection closed dialog is visible
     */
    public void connectionClosed()
    {
        // CALL method this.dispose()
        this.setVisible(false);
        // CALL method owner.endGame()
        owner.endGame();
    }

    /**
     * Shows a dialog informing user of that the connection with the server
     * could not be established.
     *
     * <br><b>preconditions:</b> none
     * <br><b>postconditions:</b> the failed connection dialog is visible
     */
    public void couldNotConnect()
    {
        // STOP the connectTimer
        connectTimer.stop();
        // CALL method this.dispose()
        this.setVisible(false);
        // CALL method owner.endGame()
        owner.endGame();
        // SHOW a message dialog box informing the user that a connection could not be made
        JOptionPane.showMessageDialog(this, "Connection could not be established.", "Connection Failed", JOptionPane.PLAIN_MESSAGE);
    }
}
