import javax.swing.*;
import javax.swing.border.*;
import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.util.StringTokenizer;
import java.util.Observer;
import java.util.jar.*;
import java.text.NumberFormat;

/**
  * The SGui class is the Graphical User Interface for the Sokoban program
  *
  * @author  Joshua Thompson
  * @date    12/01/01
  * @version 1.4
  *
  */
//Version History
//  12/01/01 Joshua Thompson added java code
//  01/12/02 Joshua Thompson modified java code; added endGame method; changed
//			 the actionPerformed method and the levelFinishedDisplay method
//  01/28/02 Joshua Thompson modified the java code; added a call to set the
//			 MapRenderer as an observer of the map in the startGame() method
//  02/24/02 Kevin Blomseth seperated GUI code into seperate class.
//  03/07/02 Kevin Blomseth added code for viewing high scores

public class SGui extends JFrame implements ActionListener
{
    /**
     * The link to the Sokoban game interface.
     */
    private Game game;
    /**
     * The link to the Tutorial.
     */
    private Tutorial tutorial;
    /**
     * Used for displaying the current map.
     */
    private MapRenderer map = new MapRenderer();
    /**
     * The main JPanel used for displaying the intro and the map.
     */
    private JPanel topPanel;
    /**
     * The text field that displays the cumulative score.
     */
    private JTextField score = new JTextField();
    /**
     * The text field that displays the current level's name.
     */
    private JTextField levelName = new JTextField();
    /**
     * The text field that displays the current number of moves.
     */
    private JTextField moves = new JTextField();
    /**
     * The text field that displays the current number of pushes.
     */
    private JTextField pushes = new JTextField();
    /**
     * The text field that displays the current number of crates incinerated.
     */
    private JTextField crates = new JTextField();
    /**
     * The text field that displays the current time in seconds.
     */
    private JTextField time = new JTextField();
    /**
     * The restart button allows the user to restart the level.
     */
    private JButton restart = new JButton("Restart");
    /**
     * The undo button allows the user to undo that last move.
     */
    private JButton undo = new JButton("Undo Move");
    /**
     * The pause button allows the user to pause the game.
     */
    private JButton pause = new JButton("Pause");
    /**
     * The end button allows the user to end the game.
     */
    private JButton end = new JButton("End Game");
    /**
     * The statsMonitor is a timer set to update the current time every second.
     */
    private Timer statsMonitor = new Timer(500, this);
    /**
     * The Intro panel to display the intro animation.
     */
    private Intro start;
    /**
     * The debug flag signifying if the game should return time accurately or
     * always 1.
     */
    protected static boolean debugMode;

    /**
     * The level editor to edit levels
     */
    private LevelEditorUI levelEditor;
    /**
     * Instance of HighScoresList
     */
    private HighScoresList highscores;

    // Declarations for the menu system
    protected JMenuBar menuBar ;
    protected JMenu gameMenu ;
    protected JMenu gameNewGame ;
    protected JMenuItem gameNewSoloGame ;
    protected JMenuItem gameNewMultiplayerGame ;
    protected JMenuItem gameLoad ;
    protected JMenuItem gameSave ;
    protected JMenu gameOptions ;
    protected JMenuItem gameOptChange ;
    protected JMenuItem gameHighScores ;
    protected JMenuItem gameLevelEditor ;
    protected JMenuItem gameEndGame ;
    protected JMenuItem gameExit ;
    protected JMenu levelMenu ;
    protected JMenuItem levelNext ;
    protected JMenuItem levelPrevious ;
    protected JMenuItem levelGoTo ;
    protected JMenuItem levelRestart ;
    protected JMenuItem levelPause ;
    protected JMenuItem levelUndo ;
    protected JMenuItem levelReplay ;
    protected JMenu helpMenu ;
    protected JMenuItem helpInstructions ;
    protected JMenuItem helpTutorial ;
    protected JMenuItem helpAbout ;

    // Declarations for the main menu buttons
    protected JButton mainSolo ;
    protected JButton mainMulti ;
    protected JButton mainHighScores ;
    protected JButton mainLevelEditor ;

    // Declarations for the end game message
    protected JButton continueButton ;
    protected JLabel levelScore ;

    // Declarations for the chat panel
    protected JPanel chatPanel ;
    protected JTextArea chatText ;
    protected JTextField chatInput ;
    protected JButton sendChat ;

    // Declarations for the Level Editor menu
    protected JMenu levelEditorLevelMenu;
    protected JMenu levelEditorHelpMenu;
    protected JMenuItem levelEditorNew;
    protected JMenuItem levelEditorLoad;
    protected JMenuItem levelEditorSave;
    protected JMenuItem levelEditorOrg;
    protected JMenuItem levelEditorExit;
    protected JMenuItem levelEditorHelpInstructions;

    /**
     * The display panel where the MapRenderer will be drawn.
     */
    protected JPanel mapDisplay ;
    /**
     * The pause screen message.
     */
    protected JLabel pauseScreen ;
    /**
     * Instance of KeyMonitor to listen to keyboard commands to control
     * the robot while the game is being played.
     */
    protected KeyMonitor keyboardListener ;


    /**
     * Constructor - initializes the window and displays the start screen
     */
    public SGui(boolean debugFlag)
    {
        // CALL the JFrame parent constructor with the window title
        super("Sokoban - Stage 3 Implementation");
        // SET the window size to Constants.normalResX x Constants.normalResY
        this.setSize(Constants.normalResX, Constants.normalResY);
        // SET the background color to light gray
        this.setBackground(Color.lightGray);
        // SET debugMode = debugFlag
        debugMode = debugFlag;
        // INIT the topPanel
        topPanel = new JPanel(new BorderLayout());
        // INIT the new Intro panel start
        start = new Intro(this);
        this.addKeyListener(start);
        // ADD start JPanel to the topPanel
        topPanel.add(start, BorderLayout.CENTER);
        // SET topPanel background to black
        topPanel.setBackground(Color.black);
        // INIT statsMonitor
        statsMonitor = new Timer(1000, this);
        // ADD the topPanel to the main window
        getContentPane().add(topPanel);
        // INIT JMenuBar system
        menuBar = new JMenuBar();
        // INIT Game menu
        gameMenu = new JMenu("Game");
        gameMenu.setMnemonic(KeyEvent.VK_G);
        gameNewGame = new JMenu("New Game");
        gameNewGame.setMnemonic(KeyEvent.VK_N);
        gameNewSoloGame = new JMenuItem("New Solo Game", KeyEvent.VK_S);
        gameNewSoloGame.setSelected(true);
        gameNewSoloGame.addActionListener(this);
        gameNewMultiplayerGame = new JMenuItem("New Multiplayer Game", KeyEvent.VK_M);
        gameNewMultiplayerGame.addActionListener(this);
        gameNewGame.add(gameNewSoloGame);
        gameNewGame.add(gameNewMultiplayerGame);
        gameLoad = new JMenuItem("Load Game", KeyEvent.VK_L);
        gameLoad.addActionListener(this);
        gameSave = new JMenuItem("Save Game", KeyEvent.VK_S);
        gameSave.addActionListener(this);
        gameOptions = new JMenu("Game Options");
        gameOptions.setMnemonic(KeyEvent.VK_O);
        gameOptChange = new JMenuItem("Change Robot Color", KeyEvent.VK_R);
        gameOptChange.addActionListener(this);
        gameOptions.add(gameOptChange);
        gameHighScores = new JMenuItem("View High Scores", KeyEvent.VK_V);
        gameHighScores.addActionListener(this);
        gameLevelEditor = new JMenuItem("Level Editor", KeyEvent.VK_E);
        gameLevelEditor.addActionListener(this);
        gameEndGame = new JMenuItem("End Game", KeyEvent.VK_E);
        gameEndGame.addActionListener(this);
        gameEndGame.setEnabled(false);
        gameEndGame.setVisible(false);
        gameExit = new JMenuItem("Exit", KeyEvent.VK_X);
        gameExit.addActionListener(this);
        gameMenu.add(gameNewGame);
        gameMenu.add(gameLoad);
        gameMenu.add(gameSave);
        gameMenu.add(gameOptions);
        gameMenu.add(gameHighScores);
        gameMenu.addSeparator();
        gameMenu.add(gameLevelEditor);
        gameMenu.add(gameEndGame);
        gameMenu.addSeparator();
        gameMenu.add(gameExit);
        // INIT Level menu
        levelMenu = new JMenu("Level");
        levelMenu.setMnemonic(KeyEvent.VK_L);
        levelNext = new JMenuItem("Next Level", KeyEvent.VK_N);
        levelNext.addActionListener(this);
        levelPrevious = new JMenuItem("Previous Level", KeyEvent.VK_P);
        levelPrevious.addActionListener(this);
        levelGoTo = new JMenuItem("Go To Level", KeyEvent.VK_G);
        levelGoTo.addActionListener(this);
        levelRestart = new JMenuItem("Restart Level", KeyEvent.VK_R);
        levelRestart.addActionListener(this);
        levelPause = new JMenuItem("Pause", KeyEvent.VK_S);
        levelPause.addActionListener(this);
        levelUndo = new JMenuItem("Undo Move", KeyEvent.VK_U);
        levelUndo.addActionListener(this);
        levelReplay = new JMenuItem("Instant Replay", KeyEvent.VK_I);
        levelReplay.addActionListener(this);
        levelMenu.add(levelNext);
        levelMenu.add(levelPrevious);
        levelMenu.add(levelGoTo);
        levelMenu.add(levelRestart);
        levelMenu.addSeparator();
        levelMenu.add(levelPause);
        levelMenu.add(levelUndo);
        levelMenu.add(levelReplay);
        // INIT Help menu
        helpMenu = new JMenu("Help");
        helpMenu.setMnemonic(KeyEvent.VK_H);
        helpInstructions = new JMenuItem("Instructions", KeyEvent.VK_I);
        helpInstructions.addActionListener(this);
        helpTutorial = new JMenuItem("Tutorial", KeyEvent.VK_T);
        helpTutorial.addActionListener(this);
        helpAbout = new JMenuItem("About", KeyEvent.VK_A);
        helpAbout.addActionListener(this);
        helpMenu.add(helpInstructions);
        helpMenu.add(helpTutorial);
        helpMenu.addSeparator();
        helpMenu.add(helpAbout);
        // ADD default menus to the menuBar system
        menuBar.add(gameMenu);
        menuBar.add(levelMenu);
        menuBar.add(helpMenu);

        // INIT chatPanel
        chatPanel = new JPanel(new BorderLayout());
        chatText = new JTextArea(5, 80);
        chatText.setLineWrap(true);
        chatInput = new JTextField();
        sendChat = new JButton("Send");
        sendChat.addActionListener(this);
        JPanel chatBottom = new JPanel(new BorderLayout());
        chatBottom.add(chatInput, BorderLayout.CENTER);
        chatBottom.add(sendChat, BorderLayout.EAST);
        JScrollPane chatScrollPane = new JScrollPane(chatText, JScrollPane.VERTICAL_SCROLLBAR_ALWAYS,
                                     JScrollPane.HORIZONTAL_SCROLLBAR_NEVER);
        chatPanel.add(chatScrollPane, BorderLayout.CENTER);
        chatPanel.add(chatBottom, BorderLayout.SOUTH);
        chatText.setEditable(false);

        // INIT pauseScreen
        pauseScreen = new JLabel("Game Paused", JLabel.CENTER);
        pauseScreen.setForeground(Color.white);
        pauseScreen.setBackground(Color.black);

        // SET undo button's ActionListener = the Sokoban class
        undo.addActionListener(this);
        undo.setMnemonic(KeyEvent.VK_U);
        // SET pause button's ActionListener = the Sokoban class
        pause.addActionListener(this);
        pause.setMnemonic(KeyEvent.VK_S);
        // SET restart button's ActionListener = the Sokoban class
        restart.addActionListener(this);
        restart.setMnemonic(KeyEvent.VK_R);
        // SET end button's ActionListener = the Sokoban class
        end.addActionListener(this);
        end.setMnemonic(KeyEvent.VK_E);
        // INIT nextLevel JButton
        continueButton = new JButton("Continue");
        continueButton.setMnemonic(KeyEvent.VK_C);
        continueButton.addActionListener(this);
        this.setDefaultCloseOperation(JFrame.DO_NOTHING_ON_CLOSE);
        // ADD a window listener to allow the X to close the program
        this.addWindowListener(new WindowAdapter ()
                               {
                                   public void windowClosing(WindowEvent e)
                                   {
                                       if (game != null)
                                       {
                                           int confirm = JOptionPane.showConfirmDialog(null, "Are you sure you " +
                                                         "want to end the current game?",
                                                         "End Game", JOptionPane.YES_NO_OPTION,
                                                         JOptionPane.PLAIN_MESSAGE);
                                           if (confirm == JOptionPane.YES_OPTION)
                                           {
                                               endGame();
                                               System.exit(0);
                                           }
                                       }
                                       else
                                       {
                                           System.exit(0);
                                       }
                                   }
                               }
                              );

        // INIT LevelEditor Level Menu
        levelEditorLevelMenu = new JMenu("Level");
        levelEditorLevelMenu.setMnemonic(KeyEvent.VK_L);
        levelEditorNew  = new JMenuItem("New Level Template", KeyEvent.VK_N);
        levelEditorNew.addActionListener(this);
        levelEditorSave = new JMenuItem("Save Level", KeyEvent.VK_S);
        levelEditorSave.addActionListener(this);
        levelEditorLoad = new JMenuItem("Load Level", KeyEvent.VK_L);
        levelEditorLoad.addActionListener(this);
        levelEditorOrg = new JMenuItem("Organize Levels", KeyEvent.VK_O);
        levelEditorOrg.addActionListener(this);
        levelEditorExit = new JMenuItem("Exit Level Editor", KeyEvent.VK_X);
        levelEditorExit.addActionListener(this);
        levelEditorLevelMenu.add(levelEditorNew);
        levelEditorLevelMenu.add(levelEditorSave);
        levelEditorLevelMenu.add(levelEditorLoad);
        levelEditorLevelMenu.add(levelEditorOrg);
        levelEditorLevelMenu.add(levelEditorExit);
        menuBar.add(levelEditorLevelMenu);

        // INIT LevelEditor Help Menu
        levelEditorHelpMenu = new JMenu("Help");
        levelEditorHelpMenu.setMnemonic(KeyEvent.VK_H);
        levelEditorHelpInstructions = new JMenuItem("Instructions", KeyEvent.VK_I);
        levelEditorHelpInstructions.addActionListener(this);
        levelEditorHelpMenu.add(levelEditorHelpInstructions);
        menuBar.add(levelEditorHelpMenu);
        levelEditorHelpMenu.setVisible(false);

        this.setName("Sokoban Main");
        // SHOW the window
        this.setVisible(true);

    }

    /**
     * Handles the events passed to the Sokoban class ActionListener.
     *
     * @param e an event passed to the ActionListener.
     * <p><b>Pre-conditions:</b>an event has occured.</p>
     * <p><b>Post-conditions:</b>the Game class has been informed of the change and the
     * user interface has been updated.</p>
     */
    public void actionPerformed (ActionEvent e)
    {
        // IF the action was triggered by the statsMonitor THEN
        if (e.getSource() == statsMonitor)
        {
            // UPDATE the stats display
            updateStats();
        }
        // ELSE IF the action was triggered by the solo game button THEN
        else if (e.getSource() == mainSolo || e.getSource() == gameNewSoloGame)
        {
            // INIT new boolean newGame = false
            boolean newGame = false;
            // IF game != null THEN
            if (game != null)
            {
                // CHECK if user wants to end the current game
                int confirm = JOptionPane.showConfirmDialog(this, "Are you sure you want" +
                              " to end the current game?",
                              "New Solo Game", JOptionPane.YES_NO_OPTION,
                              JOptionPane.PLAIN_MESSAGE);
                if (confirm == JOptionPane.YES_OPTION)
                {
                    // END the game
                    endGame();
                    newGame = true;
                }
            }
            // ELSE
            else
            {
                // SET newGame = true
                newGame = true;
            }
            // IF newGame THEN
            if (newGame)
            {
                // INIT new Game game
                game = new Game(this);
                // CALL method game.setDebugMode() passing debugMode
                game.setDebugMode(debugMode);
                // CALL method Game.startMap() passing the first level in
                //   the level list
                game.startMap(game.getFirstLevel());
                // ENABLE the buttons allowed in a solo game
                pause.setEnabled(true);
                levelEditorLevelMenu.setVisible(false);
                levelPause.setEnabled(true);
                levelReplay.setEnabled(true);
                gameSave.setEnabled(true);
                gameLoad.setEnabled(true);
                game.setMapObserver(map);
                this.requestFocus();
            }
        }
        // ELSE IF the action was triggered by the multi game button THEN
        else if (e.getSource() == mainMulti || e.getSource() == gameNewMultiplayerGame)
        {
            // INIT new boolean newGame = false
            boolean newGame = false;
            // IF game != null THEN
            if (game != null)
            {
                // CHECK if user wants to end the current game
                int confirm = JOptionPane.showConfirmDialog(this, "Are you sure you want " +
                              "to end the current game?",
                              "New Multiplayer Game",
                              JOptionPane.YES_NO_OPTION,
                              JOptionPane.PLAIN_MESSAGE);
                if (confirm == JOptionPane.YES_OPTION)
                {
                    // END current game
                    endGame();
                    newGame = true;
                }
            }
            // ELSE
            else
            {
                // SET newGame = true
                newGame = true;
            }
            // IF newGame THEN
            if (newGame)
            {
                // INIT new Game
                game = new Game(this);
                // CALL method game.setDebugMode() passing debugMode
                game.setDebugMode(debugMode);
                // SHOW the multiplayer dialog window
                MultiplayerDialog multiDiag = new MultiplayerDialog(this, game);
                // DISABLE the buttons not allowed in multiplayer
                levelEditorLevelMenu.setVisible(false);
                levelEditorHelpMenu.setVisible(false);
                pause.setEnabled(false);
                levelPause.setEnabled(false);
                levelReplay.setEnabled(false);
                gameSave.setEnabled(debugMode);
                gameLoad.setEnabled(false);
            }
        }
        // ELSE IF the action was triggered by the next level THEN
        else if (e.getSource() == levelNext)
        {
            String level = game.getNextLevel();
            if (level != null)
            {
                // RESET the stat display
                resetStatDisplay();
                // CALL method Game.changeLevel()
                game.changeLevel(level);
            }
            this.requestFocus();
        }
        // ELSE IF the action was triggered by the prev level THEN
        else if (e.getSource() == levelPrevious)
        {
            String level = game.getPrevLevel();
            if (level != null)
            {
                // RESET the stat display
                resetStatDisplay();
                // CALL method Game.changeLevel()
                game.changeLevel(level);
            }
            this.requestFocus();
        }
        // ELSE IF the action was triggered by the goto level THEN
        else if (e.getSource() == levelGoTo)
        {
            // PAUSE the game
            game.pauseTimer();
            // INIT new String level to the string value returned by showGoToDialog()
            String level = showGoToDialog();
            // IF level != null THEN
            if (level != null)
            {
                // RESET the stat display
                resetStatDisplay();
                // CALL method Game.changeLevel() passing level
                game.changeLevel(level);
            }
            this.requestFocus();
        }
        // ELSE IF the action was triggered by the continueButton THEN
        else if (e.getSource() == continueButton)
        {
            // IF game is not multiplayer or the game is hosting THEN
            if (!game.isMultiplayer() || game.isHosting())
            {
                // INIT new String nextLevel = the next level in the list
                String nextLevel = game.getNextLevel();
                // IF nextLevel != null THEN
                if (nextLevel != null)
                {
                    // resetStatDisplay()
                    resetStatDisplay();
                    // CALL method game.startMap() passing nextLevel
                    game.startMap(nextLevel);
                    // START the statsMonitor
                    statsMonitor.start();
                }
                // ELSE
                else
                {
                    // CALL method endGame()
                    endGame();
                }
            }
            this.requestFocus();
        }
        // ELSE IF the action was triggered by the restart button THEN
        else if (e.getSource() == restart || e.getSource() == levelRestart)
        {
            // RESET the stat display
            resetStatDisplay();
            // CALL method Game.restartMap()
            game.restartMap();
            this.requestFocus();
        }
        // ELSE IF the action was triggered by the undo button THEN
        else if (e.getSource() == undo || e.getSource() == levelUndo)
        {
            // CALL method Game.undoMove() passing the robot number
            Robot robot = game.getLocalRobot();
            game.undoMove(robot.getRobotNumber());
            // UPDATE stats display
            updateStats();
            this.requestFocus();
        }
        // ELSE IF the action was triggered by the instant replay menu THEN
        else if (e.getSource() == levelReplay)
        {
            // IF game is paused THEN
            if (game.isPaused())
            {
                // STOP stats monitor
                statsMonitor.stop();
                // CALL method setMapVisible()
                setMapVisible();
                // CALL game.instantReplay()
                game.instantReplay();
                // CALL method setPauseDisplay()
                setPauseDisplay();
                // START stats monitor
                statsMonitor.start();
            }
            // ELSE
            else
            {
                // STOP stats monitor
                statsMonitor.stop();
                // CALL method game.pauseTimer()
                game.pauseTimer();
                // CALL game.instantReplay()
                game.instantReplay();
                // CALL game.startTimer()
                game.startTimer();
                // START stats monitor
                statsMonitor.start();
                // ENDIF
            }
            this.requestFocus();
        }
        // ELSE IF the action was triggered by the pause button THEN
        else if (e.getSource() == pause || e.getSource() == levelPause)
        {
            // IF game is not paused THEN
            if (!game.isPaused())
            {
                // PAUSE the game
                removeKeyMonitor();
                statsMonitor.stop();
                game.pauseTimer();
                pauseGame();
                // SHOW pause display
                setPauseDisplay();
            }
            // ELSE
            else
            {
                // UNPAUSE the game
                addKeyMonitor();
                statsMonitor.start();
                game.startTimer();
                resumeGame();
                // SHOW map
                setMapVisible();
                // ENDIF
            }
            this.requestFocus();
        }
        // ELSE IF the action was triggered by the save game menu item THEN
        else if (e.getSource() == gameSave)
        {
            // PAUSE the current game
            statsMonitor.stop();
            game.pauseTimer();
            // INIT new FileDialog
            FileDialog dialog = new FileDialog(this, game);
            // CALL method dialog.showSaveGameDialog()
            dialog.showSaveGameDialog();
            // UNPAUSE the current game
            game.startTimer();
            statsMonitor.start();
            this.requestFocus();
        }
        // ELSE IF the action was triggered by the load game menu item THEN
        else if (e.getSource() == gameLoad)
        {
            // IF game != null THEN
            if (game != null)
            {
                // CHECK if the user wants to end the current game
                int confirm = JOptionPane.showConfirmDialog(this, "Loading a " +
                              "new game will end the current game.\nAre you sure?",
                              "Load Game", JOptionPane.YES_NO_OPTION,
                              JOptionPane.PLAIN_MESSAGE);
                if (confirm == JOptionPane.YES_OPTION)
                {
                    // CALL method loadGame()
                    loadGame();
                }
            }
            // ELSE
            else
            {
                // CALL method loadGame()
                loadGame();
                // ENDIF
            }
            this.requestFocus();
        }
        // ELSE IF the action was triggered by the send chat button THEN
        else if (e.getSource() == sendChat)
        {
            String message = chatInput.getText();
            String user = game.getUserName();
            chatInput.setText("");
            game.postChat(user, message);
        }
        // ELSE IF the action was triggered by the end button THEN
        else if (e.getSource() == end || e.getSource() == gameEndGame)
        {
            // CONFIRM that the user wants to end the game
            int confirm = JOptionPane.showConfirmDialog(this, "Are you sure you want" +
                          " to end the current game?", "End Game",                                                         JOptionPane.YES_NO_OPTION,
                          JOptionPane.PLAIN_MESSAGE);
            if (confirm == JOptionPane.YES_OPTION)
            {
                // CALL endGame()
                endGame();
            }
        }
        // ELSE IF the action was triggered by the change robot option THEN
        else if (e.getSource() == gameOptChange)
        {
            JOptionPane.showMessageDialog(this, "Change Robot Color: To be implemented in " +
                                          "Stage 4 Release.", "Change Robot Color",
                                          JOptionPane.PLAIN_MESSAGE);
        }
        // ELSE IF the action was triggered by the view high scores THEN
        else if (e.getSource() == gameHighScores || e.getSource() == mainHighScores)
        {
            viewHighScores();
            this.requestFocus();

        }
        // ELSE IF the action was triggered by the level editor THEN
        else if (e.getSource() == gameLevelEditor || e.getSource() == mainLevelEditor)
        {
            // Instantiate a new levelEditorUI and set it to levelEditor
            levelEditor = new LevelEditorUI();
            // CALL topPanel.removeAll to remove all other panels
            topPanel.removeAll();
            // CALL topPanel.add(levelEditor) to add the levelEditorUI
            topPanel.add(levelEditor);
            // CALL setVisible(false) on gameMenu
            gameMenu.setVisible(false);
            // CALL setVisible(false) on levelMenu
            levelMenu.setVisible(false);
            // CALL setVisible(false) on helpMenu
            helpMenu.setVisible(false);
            // CALL setVisible(true) on levelEditorLevelMenu
            levelEditorLevelMenu.setVisible(true);
            // CALL setVisible(true) on levelEditorHelpMenu
            levelEditorHelpMenu.setVisible(true);

        }
        else if (e.getSource() == levelEditorExit)
        {
            // CALL loadMainMenu()
            loadMainMenu();
            // CALL setVisible(false) on levelEditorLevelMenu
            levelEditorLevelMenu.setVisible(false);
            // CALL setVisible(false) on levelEditorHelpMenu
            levelEditorHelpMenu.setVisible(false);
            // CALL setVisible(true) on gameMenu
            gameMenu.setVisible(true);
            // CALL setVisible(true) on levelMenu
            levelMenu.setVisible(true);
            // CALL setVisible(true) on helpMenu
            helpMenu.setVisible(true);


        }
        // ELSE IF the action was triggered by level editor new level THEN
        else if (e.getSource() == levelEditorNew)
        {
            // CALL levelEditor.newLevel()
            levelEditor.newLevel();
        }
        // ELSE IF the action was triggered by level editor save level THEN
        else if (e.getSource() == levelEditorSave)
        {
            // Create a new fileChooser
            JFileChooser saveChooser = new JFileChooser();
            // SET the fileChooser to display only files
            saveChooser.setFileSelectionMode(saveChooser.FILES_ONLY);
            // SET the approve button's text
            saveChooser.setApproveButtonText("Save");
            // SET the dialog box's title
            saveChooser.setDialogTitle("Save Custom Level");
            // SET the dialog's current directory
            saveChooser.setCurrentDirectory(new File(Constants.levelPath));
            // Create a new file filter
            ExampleFileFilter filter = new ExampleFileFilter();
            // ADD "lvl" extension to the filter
            filter.addExtension("lvl");
            // SET the filter's description
            filter.setDescription("Sokoban Level Files");
            // SET the dialog's file filter to fileFilter
            saveChooser.setFileFilter(filter);
            // GET return value
            int returnVal = saveChooser.showOpenDialog(this);

            // IF the return value is the approve option THEN
            if (returnVal == JFileChooser.APPROVE_OPTION)
            {
                // Create a new File and set it to the selected file
                File file = saveChooser.getSelectedFile();
                // CALL saveLevel(file.toString())
                levelEditor.saveLevel(file.toString());
            }  // END IF
        }
        //  ELSE IF the action was triggered by the level editor load THEN
        else if (e.getSource() == levelEditorLoad)
        {
            // Create a new file chooser
            JFileChooser loadChooser = new JFileChooser();
            // SET the file chooser to display only files
            loadChooser.setFileSelectionMode(loadChooser.FILES_ONLY);
            // SET the file chooser's approve button
            loadChooser.setApproveButtonText("Load");
            // SET the file chooser's title text
            loadChooser.setDialogTitle("Load Level");
            // SET the file chooser's current directory
            loadChooser.setCurrentDirectory(new File(Constants.levelPath));
            // Create a new file filter
            ExampleFileFilter filter = new ExampleFileFilter();
            // ADD "lvl" extension to the file filter
            filter.addExtension("lvl");
            // SET filter description
            filter.setDescription("Sokoban Level Files");
            // SET file chooser's file filter
            loadChooser.setFileFilter(filter);
            // GET return value from file chooser
            int returnVal = loadChooser.showOpenDialog(this);

            // IF the return value is the approve option THEN
            if (returnVal == JFileChooser.APPROVE_OPTION)
            {
                // Create a new File and set it to the selected file
                File file = loadChooser.getSelectedFile();
                // Call loadLevel(file.toString());
                levelEditor.loadLevel(file.toString());
            }  // END IF
        }
        // ELSE IF the action was triggered by the organize levels THEN
        else if (e.getSource() == levelEditorOrg)
        {
            JOptionPane.showMessageDialog(this, "Organizing Levels: To be " +
                                          "implemented in Stage 4 Release.",
                                          "Tutorial", JOptionPane.PLAIN_MESSAGE);
        }
        // ELSE IF action triggered by level editor help
        else if (e.getSource() == levelEditorHelpInstructions)
        {
            // Display the level editor help file
            showHelpDialog(Constants.helpLevelEditorFile);
        }
        // ELSE IF the action was triggered by the help instructions THEN
        else if (e.getSource() == helpInstructions)
        {
            showHelpDialog(Constants.instructionsFile);
        }
        // ELSE IF the action was triggered by the help tutorial THEN
        else if (e.getSource() == helpTutorial)
        {
            JOptionPane.showMessageDialog(this, "Tutorial: To be implemented " +
                                          "in Stage 4 Release.", "Tutorial",
                                          JOptionPane.PLAIN_MESSAGE);
        }
        // ELSE IF the action was triggered by the help about THEN
        else if (e.getSource() == helpAbout)
        {
            JOptionPane.showMessageDialog(this, "Sokoban by Lucky Strike Software",
                                          "About Sokoban", JOptionPane.PLAIN_MESSAGE);
        }
        // ELSE IF the action was triggered by the exit menu option THEN
        else if (e.getSource() == gameExit)
        {
            // CALL System.exit() passing 0
            System.exit(0);
        }
        else if (e.getSource() == gameNewGame)
        {
            gameNewSoloGame.setSelected(true);
            // ENDIF
        }
        // REDRAW the window



        this.validate();
        this.repaint();
    }

    /**
     * Initializes the topPanel to begin displaying the map.
     *
     * @param multiplayer signifies if the game is multiplayer and the
     * chat box should be included or not
     * <p><b>Pre-conditions:</b>none.</p>
     * <p><b>Post-conditions:</b>the Game class has been initialized and the
     * game begins.</p>
     */
    public void startGame(boolean multiplayer)
    {
        // ACTIVATE the level menu
        levelMenu.setEnabled(true);
        gameSave.setEnabled(true);
        gameEndGame.setEnabled(true);
        gameEndGame.setVisible(true);
        gameLevelEditor.setEnabled(false);
        gameLevelEditor.setVisible(false);
        // CLEAR the topPanel
        topPanel.removeAll();
        // SET window size to 700 x 520
        this.setSize(Constants.normalResX, Constants.normalResY);
        // INIT new JPanel sideTop
        JPanel sideTop = new JPanel(new GridLayout(12, 1));
        // ADD a JLabel for score display
        sideTop.add(new JLabel("Score", JLabel.CENTER));
        // ADD the score JTextField
        sideTop.add(score);
        // CALL method score.setEnabled() passing false
        score.setEnabled(false);
        // ADD a JLabel for level name display
        sideTop.add(new JLabel("Level Name", JLabel.CENTER));
        // ADD the levelName JTextField
        sideTop.add(levelName);
        // CALL method levelName.setEnabled() passing false
        levelName.setEnabled(false);
        // ADD a JLabel for moves display
        sideTop.add(new JLabel("Moves", JLabel.CENTER));
        // ADD the moves JTextField
        sideTop.add(moves);
        // CALL method moves.setEnabled() passing false
        moves.setEnabled(false);
        // ADD a JLabel for pushes display
        sideTop.add(new JLabel("Pushes", JLabel.CENTER));
        // ADD the pushes JTextField
        sideTop.add(pushes);
        // CALL method pushes.setEnabled() passing false
        pushes.setEnabled(false);
        // ADD a JLabel for crates incinerated display
        sideTop.add(new JLabel("Crates Incinerated", JLabel.CENTER));
        // ADD the crates JTextField
        sideTop.add(crates);
        // CALL method crates.setEnabled() passing false
        crates.setEnabled(false);
        // ADD a JLabel for elapsed time display
        sideTop.add(new JLabel("Elapsed Time", JLabel.CENTER));
        // ADD the time JTextField
        sideTop.add(time);
        // CALL method time.setEnabled() passing false
        time.setEnabled(false);
        // INIT new JPanel sideBottom
        JPanel sideBottom = new JPanel(new GridLayout(4, 1));
        // ADD undo button
        sideBottom.add(undo);
        // ADD pause button
        sideBottom.add(pause);
        // ADD restart button
        sideBottom.add(restart);
        // ADD end button
        sideBottom.add(end);

        // INIT new JPanel sideBar
        JPanel sideBar = new JPanel(new BorderLayout());
        // SET sideBar size to 140 x Constants.normalResY
        sideBar.setSize(140, Constants.normalResY);
        // SET sideBar background color = light gray
        sideBar.setBackground(Color.lightGray);
        // ADD sideTop to the sideBar
        sideBar.add(sideTop, BorderLayout.NORTH);
        // ADD sideBottom to the sideBar
        sideBar.add(sideBottom, BorderLayout.SOUTH);
        // ADD the sideBar to the topPanel on the right
        topPanel.add(sideBar, BorderLayout.EAST);
        // INIT new MapRenderer map
        map.setOwner(game);
        // INIT new JPanel mapDisplay;
        mapDisplay = new JPanel(new BorderLayout());
        mapDisplay.setBackground(Color.black);
        if (multiplayer)
        {
            this.setSize(Constants.multiResX, Constants.multiResY);
            mapDisplay.add(chatPanel, BorderLayout.SOUTH);
            gameLoad.setEnabled(false);
            gameSave.setEnabled(debugMode);
            levelUndo.setEnabled(true);
            levelRestart.setEnabled(true);
            undo.setEnabled(true);
            restart.setEnabled(true);
        }
        //game.setMapObserver(map);
        // ADD the MapRenderer map to the mapDisplay
        mapDisplay.add(map, BorderLayout.CENTER);
        // ADD the JPanel mapDisplay to the topPanel
        topPanel.add(mapDisplay, BorderLayout.CENTER);
        // CALL method statsMonitor.start()
        statsMonitor.start();
        // REMOVE any KeyMonitors
        removeKeyMonitor();
        // INIT keyboardListener
        keyboardListener = new KeyMonitor(this, game);
        // ADD keyboardListener
        addKeyMonitor();
        // INIT stats displays
        resetStatDisplay();
        chatText.setText("");
        // REDRAW the Sokoban window
        this.requestFocus();
        if (!multiplayer)
        {
            resumeGame();
        }
        this.validate();
        this.repaint();
    }

    /**
     * Change the topPanel to display the main menu.
     *
     * <br><b>preconditions:</b> none
     * <br><b>postconditions:</b> the main menu is displayed to the user
     */
    public void loadMainMenu()
    {
        // SET the window size to Constants.normalResX x Constants.normalResY
        this.setSize(Constants.normalResX, Constants.normalResY);
        // IF menuBar not present THEN
        if (this.getJMenuBar() == null)
        {
            // SET the JMenuBar
            this.setJMenuBar(menuBar);
            // ENDIF
        }
        // DISABLE buttons not available at main menu



        levelMenu.setEnabled(false);
        gameLoad.setEnabled(true);
        gameSave.setEnabled(debugMode);
        gameEndGame.setEnabled(false);
        gameEndGame.setVisible(false);
        gameLevelEditor.setEnabled(true);
        gameLevelEditor.setVisible(true);
        levelEditorLevelMenu.setVisible(false);
        // REMOVE the current top panel
        topPanel.removeAll();
        // INIT the main menu
        JPanel mainMenu = new JPanel(new GridLayout(7, 1));
        mainMenu.setBackground(Color.black);
        // INIT mainSolo
        mainSolo = new JButton("New Solo Game");
        mainSolo.setMnemonic(KeyEvent.VK_S);
        mainSolo.addActionListener(this);
        // INIT mainMulti
        mainMulti = new JButton("New Multiplayer Game");
        mainMulti.setMnemonic(KeyEvent.VK_M);
        mainMulti.addActionListener(this);
        // INIT mainHighScores
        mainHighScores = new JButton("View High Scores");
        mainHighScores.setMnemonic(KeyEvent.VK_V);
        mainHighScores.addActionListener(this);
        // INIT mainLevelEditor
        mainLevelEditor = new JButton("Level Editor");
        mainLevelEditor.setMnemonic(KeyEvent.VK_E);
        mainLevelEditor.addActionListener(this);
        // ADD buttons to the mainMenu
        mainMenu.add(mainSolo);
        mainMenu.add(new JLabel(" "));
        mainMenu.add(mainMulti);
        mainMenu.add(new JLabel(" "));
        mainMenu.add(mainHighScores);
        mainMenu.add(new JLabel(" "));
        mainMenu.add(mainLevelEditor);
        // SET the mainMenu border
        mainMenu.setBorder(new EmptyBorder(100, 150, 100, 150));
        // ADD the mainMenu to the topPanel
        topPanel.add(mainMenu, BorderLayout.CENTER);
        // CALL method statsMonitor.stop() to stop the Timer from triggering actions
        statsMonitor.stop();
        // MAKE sure the right buttons are enabled for the next game.
        resumeGame();
    }

    /**
     * Show the load game dialog box.  If a game is loaded, change the display
     * to show the map.
     *
     * <br><b>preconditions:</b> none
     * <br><b>postconditions:</b> the go to level dialog is displayed
     */
    public void loadGame()
    {
        // INIT new Game temp = game
        Game temp = game;
        // INIT new Game
        game = new Game(this);
        // SET the debug mode = debugMode
        game.setDebugMode(debugMode);
        // INIT new FileDialog
        FileDialog dialog = new FileDialog(this, game);
        // CALL method dialog.showLoadGameDialog()
        dialog.showLoadGameDialog();
        // IF the game is loaded THEN
        if (dialog.isLoaded())
        {
            if (temp != null)
            {
                // CALL method temp.endGame();
                temp.endGame();
            }
            // CALL method startGame() passing false
            startGame(false);
            // CALL method game.setMapObserver() passing map
            game.setMapObserver(map);
        }
        // ELSE
        else
        {
            // SET game = temp
            game = temp;
        }
    }

    /**
     * Ends the game and stops the statsMonitor Timer.
     *
     * <p><b>Pre-conditions:</b>a game is currently in progress.</p>
     * <p><b>Post-conditions:</b>the statsMonitor has been stopped and the game has ended.</p>
     */
    public void endGame()
    {
        // CALL method Game.endGame()
        if (game != null)
        {
            game.endGame();
            game = null;
            this.validate();
            this.repaint();
        }
    }

    /**
     * Show the go to level dialog box.
     *
     * <br><b>preconditions:</b> none
     * <br><b>postconditions:</b> the go to level dialog is displayed
     */
    public String showGoToDialog()
    {
        // GET level list
        String[] levelNames = game.getLevelList();
        // INIT new JList levels with the level list
        JList levels = new JList(levelNames);
        levels.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
        levels.setSelectedIndex(0);
        JScrollPane levelPane = new JScrollPane(levels,
                                                JScrollPane.VERTICAL_SCROLLBAR_ALWAYS,
                                                JScrollPane.HORIZONTAL_SCROLLBAR_AS_NEEDED);
        // ADD the JList to the mainPanel
        JPanel mainPanel = new JPanel(new BorderLayout());
        mainPanel.add(levelPane, BorderLayout.CENTER);
        // DISPLAY the go to level dialog
        int confirm = JOptionPane.showConfirmDialog(this, mainPanel,
                      "Go To Level", JOptionPane.OK_CANCEL_OPTION,
                      JOptionPane.PLAIN_MESSAGE);
        // IF ok was clicked THEN
        if (confirm == JOptionPane.OK_OPTION)
        {
            // RETURN the selected level
            return (String)levels.getSelectedValue();
            // ENDIF
        }
        // RETURN null



        return null;
    }

    /**
     * Show a dialog box displaying the instructions text file.
     *
     * <br><b>preconditions:</b> none
     * <br><b>postconditions:</b> the instructions dialog is displayed
     */
    public void showHelpDialog(String helpFile)
    {
        // INIT new JPanel instructions
        JPanel help = new JPanel(new BorderLayout());
        help.setForeground(Color.black);
        // INIT JTextArea info to display text file
        JTextArea info = new JTextArea();
        info.setForeground(Color.black);
        info.setBackground(Color.white);
        info.setEditable(false);
        info.setLineWrap(true);
        info.setWrapStyleWord(true);
        JScrollPane infoScroll = new JScrollPane(info,
                                 JScrollPane.VERTICAL_SCROLLBAR_ALWAYS,
                                 JScrollPane.HORIZONTAL_SCROLLBAR_NEVER);
        infoScroll.setForeground(Color.black);
        infoScroll.setPreferredSize(new Dimension(300, 300));
        String title = "Help";
        String filename = Constants.helpPath+helpFile;
        // READ the instructions file and put into JTextArea info
        try
        {
            BufferedReader in = new BufferedReader(new FileReader(filename));
            title = in.readLine();
            while (in.ready())
            {
                info.append(in.readLine()+"\n");
            }
        }
        catch (IOException ioe)
        {
            info.setText(" Help file not found.  Make sure that the \"help\" directory has"
                         +" been extracted from the Sokoban.jar file.  To extract the"
                         +" missing files from the jar file, open a command prompt and"
                         +" change to the directory containing Sokoban.jar.  From there"
                         +" run the following command:\n\n"
                         +" jar -xf Sokoban.jar graphics levels help\n");
        }
        info.setCaretPosition(0);
        help.add(infoScroll, BorderLayout.CENTER);
        help.setBorder(new EmptyBorder(0, 0, 0, 15));
        // CALL JOptionPane.showMessageDialog() passing the instructions jpanel
        JOptionPane.showMessageDialog(this, help, title, JOptionPane.PLAIN_MESSAGE);
    }

    /**
     * The KeyMonitor class is the listener for the keyboard input from the user
     * to move the robot.
     *
     * @author  Joshua Thompson
     * @date    12/01/01
     * @version 1.0
     */
    class KeyMonitor extends KeyAdapter
    {
        /**
         * A pointer to the main Game object
         */
        Game game;
        /**
         * A pointer to the Sokoban class
         */
        SGui owner;

        public KeyMonitor (SGui owner, Game game)
        {
            // SET owner = passed owner
            this.owner = owner;
            // SET game = passed game
            this.game = game;
        }

        /**
         * Handles the keystrokes used to move the robot around the map.
         *
         * @param e an event passed to the KeyAdapter.
         * <p><b>Pre-conditions:</b>a key has been pressed.</p>
         * <p><b>Post-conditions:</b>the Game class has been informed of the move and the
         * user interface has been updated.</p>
         */
        public void keyPressed(KeyEvent ev)
        {
            // IF the key pressed is up THEN
            if (ev.getKeyCode() == KeyEvent.VK_UP)
            {
                Robot robot = game.getLocalRobot();
                // CALL method Game.postMove() passing the the robot number 0 and Direction.UP
                game.postMove(robot.getRobotNumber(), Direction.UP);
                // UPDATE stats display
                updateStats();
            }
            // ELSE IF the key pressed is down THEN
            else if (ev.getKeyCode() == KeyEvent.VK_DOWN)
            {
                Robot robot = game.getLocalRobot();
                // CALL method Game.postMove() passing the robot number 0 and Direction.DOWN
                game.postMove(robot.getRobotNumber(), Direction.DOWN);
                // UPDATE stats display
                updateStats();
            }
            // ELSE IF the key pressed is left THEN
            else if (ev.getKeyCode() == KeyEvent.VK_LEFT)
            {
                Robot robot = game.getLocalRobot();
                // CALL method Game.postMove() passing the robot number 0 and Direction.LEFT
                game.postMove(robot.getRobotNumber(), Direction.LEFT);
                // UPDATE stats display
                updateStats();
            }
            // ELSE IF the key pressed is right THEN
            else if (ev.getKeyCode() == KeyEvent.VK_RIGHT)
            {
                Robot robot = game.getLocalRobot();
                // CALL method Game.postMove() passing the robot number 0 and Direction.RIGHT
                game.postMove(robot.getRobotNumber(), Direction.RIGHT);
                // UPDATE stats display
                updateStats();
            }
            // REDRAW the Sokoban window
            owner.validate();
            owner.repaint();
        }
    } // end class KeyMonitor

    /**
     * Adds a KeyMonitor
     */
    public void addKeyMonitor()
    {
        // GET the listeners
        java.util.EventListener[] listeners = this.getListeners(KeyListener.class);
        // IF no listeners THEN
        if (listeners.length == 0)
        {
            // ADD a KeyListener new KeyMonitor
            this.addKeyListener(keyboardListener);
            // ENDIF
        }
    }

    /**
     * Removes all KeyMonitors
     */
    public void removeKeyMonitor()
    {
        // GET the listeners
        java.util.EventListener[] listeners = this.getListeners(KeyListener.class);
        // FOR index = 0 to listeners.length
        for (int index = 0; index < listeners.length; index++)
        {
            // ADD a KeyListener new KeyMonitor
            this.removeKeyListener((KeyMonitor)listeners[index]);
            // ENDFOR
        }
    }

    /**
     * Displays the level finished screen giving the user's game statistics
     *
     * @param levelScore the score received for finishing the level
     * <p><b>Pre-conditions:</b>a level has been finished.</p>
     * <p><b>Post-conditions:</b>the level finished screen is displayed to the user.</p>
     */
    public void setLevelFinishedDisplay(int levelScore)
    {
        if (mapDisplay != null)
        {
            removeKeyMonitor();
            pause.setEnabled(false);
            levelPause.setEnabled(false);
            restart.setEnabled(false);
            levelRestart.setEnabled(false);
            levelReplay.setEnabled(false);
            undo.setEnabled(false);
            levelUndo.setEnabled(false);
            mapDisplay.remove(map);
            // GET the user's robot
            Robot robot = game.getLocalRobot();
            // INIT new JPanel stats with a GridLayout of 10 x 2
            JPanel stats = new JPanel(new GridLayout(10, 2));
            stats.setBackground(Color.black);
            stats.setForeground(Color.white);
            JLabel movesHeader = new JLabel("Moves:");
            movesHeader.setForeground(Color.white);
            // ADD new JLabel for moves display
            stats.add(movesHeader);
            JLabel movesDisp = new JLabel("   "+robot.getMoves());
            movesDisp.setForeground(Color.white);
            // ADD new JLabel for actual number of moves
            stats.add(movesDisp);
            JLabel pushesHeader = new JLabel("Pushes:");
            pushesHeader.setForeground(Color.white);
            // ADD new JLabel for pushes display
            stats.add(pushesHeader);
            JLabel pushesDisp = new JLabel("   "+robot.getPushes());
            pushesDisp.setForeground(Color.white);
            // ADD new JLabel for actual number of pushes
            stats.add(pushesDisp);
            JLabel timeHeader = new JLabel("Time:");
            timeHeader.setForeground(Color.white);
            // ADD new JLabel for time taken display
            stats.add(timeHeader);
            // ADD new JLabel for actual time taken
            NumberFormat timeFormat = NumberFormat.getInstance();
            timeFormat.setMinimumIntegerDigits(2);
            int minutes = game.getTime()/60;
            int seconds = game.getTime()%60;
            String timeStr = timeFormat.format(minutes)+":"+timeFormat.format(seconds);
            JLabel timeLabel = new JLabel("   "+timeStr);
            timeLabel.setForeground(Color.white);
            stats.add(timeLabel);
            JLabel levelScoreHeader = new JLabel("Score");
            levelScoreHeader.setForeground(Color.white);
            // ADD new JLabel for score display
            stats.add(levelScoreHeader);
            // ADD new JLabel for actual level score
            this.levelScore = new JLabel("   "+levelScore);
            this.levelScore.setForeground(Color.white);
            stats.add(this.levelScore);
            JLabel totalScoreHeader = new JLabel("Total Score:");
            totalScoreHeader.setForeground(Color.white);
            // ADD new JLabel for total score display
            stats.add(totalScoreHeader);
            JLabel totalScore = new JLabel("   "+game.getTotalScore());
            totalScore.setForeground(Color.white);
            // ADD new JLabel for actual total score
            stats.add(totalScore);
            // SET border for stats
            stats.setBorder(new EmptyBorder(10, 0, 10, 0));
            // INIT new JPanel buttons with FlowLayout
            JPanel buttons = new JPanel(new FlowLayout());
            // ADD continueButton to buttons
            buttons.add(continueButton);
            continueButton.setDefaultCapable(true);
            continueButton.requestFocus();
            this.getRootPane().setDefaultButton(continueButton);
            // SET buttons background = Color.black
            buttons.setBackground(Color.black);
            buttons.setForeground(Color.white);
            // INIT new JPanel display with BorderLayout
            JPanel display = new JPanel(new BorderLayout());
            JLabel finished = new JLabel("Level Finished", JLabel.CENTER);
            finished.setForeground(Color.white);
            if (game.getNextLevel() == null)
            {
                finished.setText("Game Over");
                continueButton.setText("End Game");
            }
            else
            {
                continueButton.setText("Continue");
                finished.setText("Level Finished");
            }
            // ADD game summary title JLabel
            display.add(finished, BorderLayout.NORTH);
            // ADD the stats JPanel
            display.add(stats, BorderLayout.CENTER);
            if (game.isClientConnected())
            {
                JLabel waiting = new JLabel("Waiting for host...", JLabel.CENTER);
                waiting.setForeground(Color.white);
                display.add(waiting, BorderLayout.SOUTH);
            }
            else
            {
                // ADD the buttons JPanel
                display.add(buttons, BorderLayout.SOUTH);
            }
            // SET display border
            display.setBorder(new EmptyBorder(50, 0, 0, 0));
            display.setBackground(Color.black);
            display.setForeground(Color.white);
            // SET score text field to reflect new total score
            score.setText(""+game.getTotalScore());
            // INIT new showStats JPanel with FlowLayout
            JPanel showStats = new JPanel(new FlowLayout());
            // ADD display to showStats
            showStats.add(display);
            // SET showStats background = Color.black
            showStats.setBackground(Color.black);
            showStats.setForeground(Color.white);
            statsMonitor.stop();
            mapDisplay.add(showStats, BorderLayout.CENTER);
        }
        this.validate();
        this.repaint();
    }

    /**
     * Disable the UI buttons not available during a pause game.
     *
     * <br><b>preconditions:</b> none
     * <br><b>postconditions:</b> the UI buttons not available when a game
     * is paused are disabled
     */
    public void pauseGame()
    {
        // CHANGE the text of the pause buttons to "Resume"
        pause.setText("Resume");
        levelPause.setText("Resume");
        // DISABLE the buttons not available when the game is paused
        levelNext.setEnabled(false);
        levelPrevious.setEnabled(false);
        levelGoTo.setEnabled(false);
        levelUndo.setEnabled(false);
        undo.setEnabled(false);
        levelRestart.setEnabled(false);
        restart.setEnabled(false);
    }

    /**
     * Enable the UI buttons disabled after a pause game.
     *
     * <br><b>preconditions:</b> none
     * <br><b>postconditions:</b> the UI buttons not available when a game
     * is paused are enabled
     */
    public void resumeGame()
    {
        // CHANGE the pause buttons text to "Pause"
        pause.setText("Pause");
        levelPause.setText("Pause");
        // ENABLE the buttons not available when the game is paused
        pause.setEnabled(true);
        levelPause.setEnabled(true);
        levelNext.setEnabled(true);
        levelPrevious.setEnabled(true);
        levelGoTo.setEnabled(true);
        levelUndo.setEnabled(true);
        undo.setEnabled(true);
        levelReplay.setEnabled(true);
        levelRestart.setEnabled(true);
        restart.setEnabled(true);
    }

    /**
     * Updates the stat display.
     *
     * <br><b>preconditions:</b> the game is running
     * <br><b>postconditions:</b> the stats are updated to reflect the current
     * state of the game
     */
    public void updateStats()
    {
        // IF game != null THEN
        if (game != null)
        {
            // INIT new robot = game.getLocalRobot()
            Robot robot = game.getLocalRobot();
            // SET levelName = game.getCurrentLevel()
            levelName.setText(game.getCurrentLevel());
            // SET moves text = robot.getMoves()
            moves.setText(""+robot.getMoves());
            // SET pushes text = robot.getPushes()
            pushes.setText(""+robot.getPushes());
            // SET crates text = robot.getCratesIncinerated()
            crates.setText(""+robot.getCratesIncinerated());
            // INIT new int curTime
            int curTime = game.getTime();
            // INIT minutes = curTime / 60
            int minutes = curTime / 60;
            // INIT seconds = curTime % 60
            int seconds = curTime % 60;
            // INIT new NumberFormat
            NumberFormat timeFormat = NumberFormat.getInstance();
            // SET minimum integer digits = 2
            timeFormat.setMinimumIntegerDigits(2);
            // INIT timeStr = minutes:seconds
            String timeStr = timeFormat.format(minutes)+":"+timeFormat.format(seconds);
            // SET time text = timeStr
            time.setText(timeStr);
            // ENDIF
        }
    }

    /**
     * Resets the stat display.
     *
     * <br><b>preconditions:</b> none
     * <br><b>postconditions:</b> the stats are set to how they were at the
     * beginning of the game
     */
    public void resetStatDisplay()
    {
        score.setText(""+game.getTotalScore());
        levelName.setText(game.getCurrentLevel());
        moves.setText("0");
        pushes.setText("0");
        crates.setText("0");
        time.setText("00:00");
    }

    /**
     * Change the game display to show the pause screen and the map.
     *
     * <br><b>preconditions:</b> the map is being displayed
     * <br><b>postconditions:</b> the map is removed and the pause screen is
     * being displayed
     */
    public void setPauseDisplay()
    {
        if (mapDisplay != null)
        {
            mapDisplay.remove(map);
            mapDisplay.add(pauseScreen, BorderLayout.CENTER);
        }
        this.validate();
        this.repaint();
    }

    /**
     * Change the game display to show the map.
     *
     * <br><b>preconditions:</b> none
     * <br><b>postconditions:</b> the map is displayed
     */
    public void setMapVisible()
    {
        if (mapDisplay != null)
        {
            mapDisplay.remove(mapDisplay.getComponentAt(1,1));
            mapDisplay.add(map, BorderLayout.CENTER);
        }
        this.validate();
        this.repaint();
    }

    /**
     * Changes the level name displayed to the user.
     *
     * @param name the new level name to display to the user
     * <p><b>Pre-conditions:</b>the level has been changed.</p>
     * <p><b>Post-conditions:</b>the level name displayed to the user is updated.</p>
     */
    public void setLevelName(String name)
    {
        levelName.setText(name);
    }

    /**
     * Displays the passed text in the chat window
     */
    public void addText(String str)
    {
        chatText.append(str+"\n");
        chatText.setCaretPosition(chatText.getText().length());
    }

    /**
     * Displays the High Scores window
     */
    protected void viewHighScores()
    {
        highscores = new HighScoresList(game);
        highscores.loadHighScores();
        HighScore[] scores = highscores.toArray();
        // INIT new JPanel
        JPanel scoreDisplay = new JPanel(new BorderLayout());
        scoreDisplay.setForeground(Color.black);
        scoreDisplay.setBackground(Color.lightGray);
        // INIT JTextArea info to display scores
        JTextArea scoreArea = new JTextArea();
        scoreArea.setForeground(Color.black);
        scoreArea.setBackground(Color.lightGray);
        scoreArea.setEditable(false);
        scoreArea.setRows(16);
        scoreDisplay.add(scoreArea);
        String title = "High Scores";
        for(int index = 0; index < scores.length; index++)
        {
            scoreArea.append((index+1) + " " + scores[index].getName() +
                             " " + scores[index].getScore()+"\n");
        }
        scoreDisplay.setBorder(new EmptyBorder(0, 0, 0, 15));
        // CALL JOptionPane.showMessageDialog() passing the instructions jpanel
        JOptionPane.showMessageDialog(this, scoreDisplay, title, JOptionPane.PLAIN_MESSAGE);
    }
}


