import java.io.*;
import java.net.*;
import java.util.*;

/**
 * <p>The Server class is a communications interface for an individual user who
 * wants to host a multiplayer game of Sokoban.</p>
 *
 * @author Joshua Thompson
 * @date 11/23/2001
 * @version 1.2
 *
 */
//Version History
//  11/23/01 Joshua Thompson added algorithm pseudocode
//  11/28/01 Joshua Thompson modified javadoc headers and added variable
//			 descriptions

public class Server extends Thread
{

    /**
     * A pointer to the main Game object
     */
    private Game owner;
    /**
     * the maximum number of client connections allowed
     */
    private int maxCon = 3;
    /**
     * the port to listen for connections on
     */
    private int port;
    /**
     * accepts connections from a client
     */
    private ServerSocket serverSocket;
    /** The vector clients will be a vector of clients, with a client object
     *  for each person connected to the server.
     */
    private Vector clients;
    /**
     * The total number of connections currently active
     */
    private int con;
    /**
     * the current thread number used for identifying clients
     */
    private int threadNum;
    /**
     * whether or not the server is listening for client connections
     */
    private boolean listening;

    /**
     * <p>Initializes Server and waits for connections.</p>
     *
     * <p><b>Pre-condition:</b> none</p>
     * <p><b>Post-condition:</b> server is waiting for connections</p>
     */
    public void run()
    {
        // INIT serverSocket as new ServerSocket listening on the stored port
        try
        {
            serverSocket = new ServerSocket(port);
        }
        catch (IOException ioe)
        {
            ioe.printStackTrace();
        }
        // SET listening = true
        listening = true;
        // SET con = 0
        con = 0;
        // SET threadNum = 1
        threadNum = 1;
        // WHILE listening LOOP
        while (listening)
        {
            try
            {
                // CALL method accept() from serverSocket
                Socket clientSocket = serverSocket.accept();
                // IF con is less than maxCon THEN
                if (con < maxCon)
                {
                    // CALL method sendNewUser()
                    // INIT new ServerThread with owner, clientSocket,and threadNum
                    ServerThread st = new ServerThread(owner, clientSocket, threadNum);
                    // ADD created ServerThread to end of clients Vector
                    clients.addElement(st);
                    // START created ServerThread
                    st.start();
                    // INCREMENT con
                    con++;
                    // INCREMENT threadNum
                    threadNum++;
                    // ENDIF
                }
                else
                {
                    PrintWriter out = new PrintWriter(clientSocket.getOutputStream(), true);
                    out.println("-1 Too many connections.");
                    out.close();
                    clientSocket.close();
                }
            }
            catch (IOException ioe)
            {
                ioe.printStackTrace();
            }
            // END WHILE
        }
        // CLOSE the serverSocket

        try
        {
            serverSocket.close();
        }
        catch (IOException ioe)
        {
            ioe.printStackTrace();
        }
    }

    /**
     * <p>Creates a move command String for the robot passed and calls the
     * sendToAll() method to send the command to all connected clients.</p>
     *
     * <p><b>Pre-condition:</b> none</p>
     * <p><b>Post-condition:</b> a move command String is created and sent to the
     * sendToAll() method.</p>
     *
     * @param robot the robot number of the robot moving
     * @param dir the direction the robot is moving
     */
    public void sendMove(int robot, Direction dir)
    {
        // CREATE new String with move keyword, robot, and direction
        String command = "move " + robot + " " + dir.toString();
        // CALL method sendToAll with created String as parameter
        this.sendToAll(command);
    }

    /**
     * <p>Creates an undo command String for the robot passed and calls the
     * sendToAll() method to send the command to all connected clients.</p>
     *
     * <p><b>Pre-condition:</b> none</p>
     * <p><b>Post-condition:</b> an undo command String is created and sent to the
     * sendToAll() method.</p>
     *
     * @param robot the robot number of the robot moving
     */
    public void sendUndo(int robot)
    {
        // CREATE new String with undo keyword and robot number passed
        String command = "undo " + robot;
        // CALL method sendToAll with created String as parameter
        this.sendToAll(command);
    }

    /**
     * <p>Creates a disconnect command String for the robot passed and calls the
     * sendToAll() method to send the command to all connected clients.</p>
     *
     * <p><b>Pre-condition:</b> none</p>
     * <p><b>Post-condition:</b> a disconnect command String is created and
     * sent to the sendToAll() method.</p>
     *
     * @param robot the robot number of the robot moving
     */
    public void sendDisconnect(int robot)
    {
        // DECREMENT con
        this.con--;
        // CREATE new String with disconnect keyword and robot number passed
        String command = "disconnect " + robot;
        // CALL method sendToAll with created String as parameter
        this.sendToAll(command);
        // INIT new int index = 0
        int index = 0;
        // INIT new int threadNum
        int threadNum;
        // WHILE index < clients.size() DO
        while (index < clients.size())
        {
            // SET threadNum = clients[index].getThreadNum()
            threadNum = ((ServerThread)clients.elementAt(index)).getThreadNum();
            // IF threadNum = robot
            if (threadNum == robot)
            {
                // REMOVE ServerThread from clients vector
                clients.removeElementAt(index);
                // SET index = clients.size() to stop loop
                index = clients.size();
            }
            // ELSE
            else
            {
                // INCREMENT index
                index++;
                // ENDIF
            }
            // ENDWHILE
        }
    }


    /**
     * <p>Creates a chat command String for the robot passed and calls the
     * sendToAll() method to send the command to all connected clients.</p>
     *
     * <p><b>Pre-condition:</b> none</p>
     * <p><b>Post-condition:</b> a chat command String is created and sent to the
     * sendToAll() method.</p>
     *
     * @param user the user name of the user who sent the message
     * @param message the message sent by the user
     */
    public void sendChat(String user, String message)
    {
        // CREATE new String with chat keyword, user name and message
        String command = "chat " + user + " " + message;
        // CALL method sendToAll with created String as parameter
        this.sendToAll(command);
    }

    /**
     * <p>Creates a restartMap command String for the robot passed and calls the
     * sendToAll() method to send the command to all connected clients.</p>
     *
     * <p><b>Pre-condition:</b> none</p>
     * <p><b>Post-condition:</b> a restart command String is created and sent to the
     * sendToAll() method.</p>
     */
    public void sendRestartMap()
    {
        // CREATE new String with restartMap keyword
        String command = "restart";
        // CALL method sendToAll with created String as parameter
        this.sendToAll(command);
    }

    /**
     * <p>Creates a restart request command String for the robot passed and
     * calls the sendToAll() method to send the command to all connected
     * clients.</p>
     *
     * <p><b>Pre-condition:</b> none</p>
     * <p><b>Post-condition:</b> a restart command String is created and sent to the
     * sendToAll() method.</p>
     */
    public void sendRestartRequest(int robotNum)
    {
        // CREATE new String with restartMap keyword
        String command = "resreq " + robotNum;
        // CALL method sendToAll with created String as parameter
        this.sendToAll(command);
    }

    /**
     * <p>Creates a restart request command String for the robot passed and
     * calls the sendToAll() method to send the command to all connected
     * clients.</p>
     *
     * <p><b>Pre-condition:</b> none</p>
     * <p><b>Post-condition:</b> a restart command String is created and sent to the
     * sendToAll() method.</p>
     */
    public void sendChangeLevelRequest(String level, int robotNum)
    {
        // CREATE new String with restartMap keyword
        String command = "chgreq " + robotNum + " " + level;
        // CALL method sendToAll with created String as parameter
        this.sendToAll(command);
    }

    /**
     * <p>Sends a command String to all connected clients.</p>
     *
     * <p><b>Pre-condition:</b> the command parameter is a properly formatted
     * String object.</p>
     * <p><b>Post-condition:</b> the command parameter is sent to each
     * ServerThread to be sent to each client.</p>
     *
     * @param command a formatted command String to be sent to each client connected
     */
    protected void sendToAll(String command)
    {
        // FOR index = 0 to end of clients vector
        for (int index = 0; index < clients.size(); index++)
        {
            // CALL method sendCommand() from clients[index] passing the command String
            ((ServerThread)clients.elementAt(index)).sendCommand(command);
            // END FOR
        }
    }

    /**
     * <p>Sends the current Map to all clients.</p>
     *
     * <p><b>Pre-condition:</b> server has a map loaded in the board[][] in Map class</p>
     * <p><b>Post-condition:</b> all clients' Map classes are updated.</p>
     */
    public void sendMap(String level)
    {
        // CALL getMap() from owner and store in tempMap
        String[] map = owner.getMapString();
        // INIT String command with the map keyword and the level name
        String command = "map "+ level;
        // FOR index = 0 to end of clients vector
        for (int index = 0; index < clients.size(); index++)
        {
            // CALL method sendCommand() from client[index] passing the command String
            ((ServerThread)clients.elementAt(index)).sendCommand(command);
            // FOR index = 0 to end of row
            for (int row = 0; row < map.length; row++)
            {
                // CALL method sendCommand() from client[index] passing tempMap[row]
                ((ServerThread)clients.elementAt(index)).sendCommand(map[row]);
                // ENDFOR
            }
            // ENDFOR
        }
    }

    /**
     * <p>Creates an add user command String for the robot passed and
     * calls the sendToAll() method to send the command to all connected
     * clients.</p>
     *
     * <p><b>Pre-condition:</b> none</p>
     * <p><b>Post-condition:</b> an add user command String is created and
     * sent to the sendToAll() method.</p>
     */
    public void addUser(String userName, int robotNum)
    {
        // INIT command String
        String command = "adduser " + userName + " " + robotNum;
        // CALL method sendToAll() passing command
        sendToAll(command);
    }

    /**
     * <p>Stops the server and sends a disconnect message to each client.</p>
     *
     * <p><b>Pre-condition:</b> existing socket connection to a server</p>
     * <p><b>Post-condition:</b> disconnect command sent to all clients;
     * server stoped listening for connections</p>
     */
    public void stopServer()
    {
        // FOR index = 0 to end of clients vector
        for (int index = 0; index < clients.size(); index++)
        {
            // CREATE new command String with the disconnect keyword and
            //    client[index].getThreadNum()
            String command = "disconnect " + ((ServerThread)clients.elementAt(index)).getThreadNum();
            // CALL method sendCommand() from clients[index] passing the
            //    command String
            ((ServerThread)clients.elementAt(index)).sendCommand(command);
            // END FOR
        }
        while (owner.getRobots().length > 1)
        {
        }
        // CALL method Server.stopListening()

        if (listening)
        {
            stopListening();
        }
        //listening = false;
    }

    /**
     * <p>Server constructor</p>
     *
     * <p><b>Pre-condition:</b> none</p>
     * <p><b>Post-condition:</b> Server class created; the internal owner variable is set to passed value.</p>
     *
     * @param owner provides a link back to the game.
     */
    public Server(Game owner, int port)
    {
        // SET owner = passed owner
        this.owner = owner;
        // SET port = passed port
        this.port = port;
        // INIT clients Vector
        this.clients = new Vector();
    }

    /**
     * <p>Sets the port to listen on.</p>
     *
     * @param port the port to listen on
     */
    public void setPort(int port)
    {
        // SET port to passed port
        this.port = port;
    }

    /**
     * <p>Returns the number of the port to listen on.</p>
     *
     * @return port - the port used to listen for connections
     */
    public int getPort()
    {
        // RETURN port
        return port;
    }

    /**
     * <p>Sets the maximum number of connections allowed.</p>
     *
     * @param max the maximum number of connections allowed
     */
    public void setMaxCon(int max)
    {
        // SET maxCon to max
        this.maxCon = max;
    }

    /**
     * <p>Returns the maximum number of connections allowed.</p>
     *
     * @return maxCon - the maximum number of connections to listen for
     */
    public int getMaxCon()
    {
        // RETURN maxCon
        return maxCon;
    }

    /**
     * <p>Returns the thread numbers of each ServerThread in the clients
     * vector.</p>
     *
     * @return the thread numbers of each ServerThread
     */
    public int[] getThreadNums()
    {
        // INIT int[] threadNums
        int[] threadNums = new int[clients.size()+1];
        // SET threadNums[0] = 0;
        threadNums[0] = 0;
        // FOR currentThread = 0 to end of clients vector
        for (int currentThread = 0; currentThread < clients.size(); currentThread++)
        {
            // SET threadNums[currentThread+1] = clients.getThreadNum()
            threadNums[currentThread+1] = ((ServerThread)clients.elementAt(currentThread)).getThreadNum();
            // ENDFOR
        }
        // RETURN threadNums

        return threadNums;
    }

    /**
     * Stops the Server from listening for connections.
     *
     * <br><b>preconditions:</b> the server is listening for connections.
     * <br><b>postconditions:</b> the server has stopped listening for connections.
     */
    public void stopListening()
    {
        // IF listening THEN
        if (listening)
        {
            // SET listening = false
            listening = false;
            // STOP listening for connections
            try
            {
                Socket killServ = new Socket("localhost", getPort());
                if (clients.size() < maxCon)
                {
                    PrintWriter kOut = new PrintWriter(killServ.getOutputStream(), true);
                    BufferedReader kIn = new BufferedReader(new InputStreamReader(killServ.getInputStream()));
                    kOut.println("killserv");
                    kOut.println("disconnect");
                    kOut.println("disconnect");
                    kIn.readLine();
                    kIn.readLine();
                    kIn.readLine();
                    kOut.close();
                    kIn.close();
                }
                killServ.close();
            }
            catch (IOException ioe)
            {
                ioe.printStackTrace();
                System.out.println("could not kill server");
            }
            // ENDIF
        }
    }
}
